"""
SPICE training pipeline as a scikit-learn estimator
"""
import warnings
import time
import torch
import numpy as np
from copy import deepcopy
from sklearn.base import BaseEstimator
from typing import Dict, Optional, Tuple, Union, Iterable, List

from .resources.rnn_training import fit_model
from .resources.rnn_utils import load_checkpoint, DatasetRNN
from .resources.sindy_utils import load_spice, save_spice, check_library_setup
from .resources.sindy_training import fit_spice
from .resources.bandits import AgentNetwork, AgentSpice
from .resources.rnn import BaseRNN, ParameterModule


warnings.filterwarnings("ignore")

class SpiceConfig():
    def __init__(self,
                 library_setup: Dict[str, Iterable[str]],
                 filter_setup: Dict[str, Iterable[Union[str, float, int, bool]]],
                 control_parameters: Iterable[str],
                 rnn_modules: Iterable[str]):
        """
        Config class for SPICE model.

        Args:
            library_setup: Dictionary of library setup
            filter_setup: Dictionary of filter setup
            control_parameters: List of control parameters
            rnn_modules: List of RNN modules
        """
        self.library_setup = library_setup
        self.filter_setup = filter_setup
        self.control_parameters = control_parameters
        self.rnn_modules = rnn_modules

        self.spice_feature_list = rnn_modules + control_parameters

        if not check_library_setup(self.library_setup, self.spice_feature_list, verbose=True):
            raise ValueError('\nLibrary setup does not match feature list.')


class SpiceEstimator(BaseEstimator):
    """
    Scikit-learn estimator for fitting the SPICE model and making predictions.
    
    Combines an RNN for predicting behavioral choices with SPICE for discovering
    the underlying dynamical equations of cognitive mechanisms.
    """
    
    def __init__(
        self,
        
        # RNN class and SPICE configuration. Can be one of the precoded models in rnn.py or a custom implementation.
        rnn_class: BaseRNN,
        spice_config: SpiceConfig,
        
        # Data/Environment parameters
        n_actions: int = 2,
        n_participants: int = 0,
        n_experiments: int = 0,
        
        # RNN training parameters
        epochs: int = 128,
        bagging: bool = False,
        sequence_length: Optional[int] = -1,  # -1 for keeping the sequence length in the data to its original length, otherwise strided windows of length sequence_length,
        n_steps_per_call: Optional[int] = -1,  # number of timesteps in one backward-call; -1 for full sequence
        batch_size: Optional[int] = -1,  # -1 for a batch-size equal to the number of participants in the data
        learning_rate: Optional[float] = 1e-2,
        convergence_threshold: Optional[float] = 0,
        device: Optional[torch.device] = torch.device('cpu'),
        scheduler: Optional[bool] = False, 
        train_test_ratio: Optional[Union[float, List[int]]] = 1.,
        l2_weight_decay: Optional[float] = 1e-4,
        dropout: Optional[float] = 0.5,
        
        # SPICE training parameters
        spice_optimizer_type: Optional[str] = 'SR3_weighted_l1',
        spice_optim_threshold: Optional[float] = 0.05,
        spice_optim_regularization: Optional[float] = 1e-2,
        spice_library_polynomial_degree: Optional[int] = 1,
        n_trials_off_policy: Optional[int] = 1000,
        n_sessions_off_policy: Optional[int] = 1,
        n_trials_same_action_off_policy: Optional[int] = 5,
        use_optuna: Optional[bool] = False,
        optuna_threshold: Optional[float] = 0.1,
        optuna_n_trials: Optional[int] = 50,
        
        verbose: Optional[bool] = False,

        save_path_rnn: Optional[str] = None,
        save_path_spice: Optional[str] = None
    ):
        """
        Args:
            rnn_class: RNN class. Can be one of the precoded models in rnn.py or a custom implementation.
            spice_config: SPICE config
            list_signals: List of signals to use for SPICE
            hidden_size: Hidden size of the RNN
            dropout: Dropout rate of the RNN
            n_actions: Number of actions
            n_participants: Number of participants
            n_experiments: Number of experiments
            epochs: Number of epochs to train the RNN
            bagging: Whether to use bagging for the RNN
            sequence_length: Sequence length for the RNN
            n_steps_per_call: Number of steps per call for the RNN
            batch_size: Batch size for the RNN
            learning_rate: Learning rate for the RNN
            convergence_threshold: Convergence threshold for the RNN
            device: Device to use for the RNN (default: 'cpu')
            scheduler: Whether to use a scheduler for the RNN (default: False)
            train_test_ratio: Ratio of training to test data (default: 1.)
            l2_weight_decay: L2 weight decay for the RNN
            verbose: Whether to print verbose output (default: False)
            save_path_rnn: File path (.pkl) to save RNN model after training (default: None)
            save_path_spice: File path (.pkl) to save SPICE model after training (default: None)
        """
        
        super(BaseEstimator, self).__init__()
        
        # Training parameters
        self.epochs = epochs
        self.bagging = bagging
        self.sequence_length = sequence_length
        self.n_steps_per_call = n_steps_per_call
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.convergence_threshold = convergence_threshold
        self.scheduler = scheduler
        self.train_test_ratio = train_test_ratio
        self.device = device
        self.verbose = verbose
        self.deterministic = False
        self.l2_weight_decay = l2_weight_decay

        # Save parameters
        self.save_path_rnn = save_path_rnn
        self.save_path_spice = save_path_spice

        # SPICE training parameters
        self.spice_optimizer_type = spice_optimizer_type
        self.spice_optim_threshold = spice_optim_threshold
        self.spice_library_polynomial_degree = spice_library_polynomial_degree
        self.spice_optim_regularization = spice_optim_regularization
        self.n_trials_off_policy = n_trials_off_policy
        self.n_sessions_off_policy = n_sessions_off_policy
        self.n_trials_same_action_off_policy = n_trials_same_action_off_policy
        self.use_optuna = use_optuna
        self.optuna_threshold = optuna_threshold
        self.optuna_n_trials = optuna_n_trials

        # Data parameters
        self.n_actions = n_actions
        self.n_participants = n_participants
        self.n_experiments = n_experiments
        
        # RNN parameters
        self.dropout = dropout
        
        self.rnn_agent = None
        self.spice_agent = None
        self.spice_features = None
        
        self.rnn_model = rnn_class(
            n_actions=n_actions,
            n_participants=n_participants,
            n_experiments=n_experiments,
        ).to(device)

        self.spice_library_config = spice_config.library_setup
        self.spice_filter_config = spice_config.filter_setup
        self.control_parameters = spice_config.control_parameters
        self.rnn_modules = spice_config.rnn_modules

        self.spice_feature_list = spice_config.spice_feature_list

        self.optimizer_rnn = torch.optim.Adam(self.rnn_model.parameters(), lr=learning_rate)
    
    def fit(self, data: np.ndarray, targets: np.ndarray, data_test: np.ndarray = None, target_test: np.ndarray = None):
        """
        Fit the RNN and SPICE models to given data.
        
        Args:
            conditions: Array of shape (n_participants, n_trials, n_features)
            targets: Array of shape (n_participants, n_trials, n_actions)
        """
        
        dataset = DatasetRNN(data, targets)
        dataset_test = DatasetRNN(data_test, target_test) if data_test is not None and target_test is not None else None
        
        start_time = time.time()
        
        # ------------------------------------------------------------------------
        # Fit RNN
        # ------------------------------------------------------------------------
        
        # if self.verbose:
        print('\nTraining the RNN...')
        
        batch_size = data.shape[0] if self.batch_size == -1 else self.batch_size
        
        rnn_model, rnn_optimizer, _ = fit_model(
            model=self.rnn_model,
            dataset_train=dataset,
            dataset_test=dataset_test,
            optimizer=self.optimizer_rnn,
            convergence_threshold=self.convergence_threshold,
            epochs=self.epochs,
            batch_size=batch_size,
            bagging=self.bagging,
            scheduler=self.scheduler,
            n_steps=self.n_steps_per_call,
            verbose=self.verbose,
            path_save_checkpoints=None,
        )

        rnn_model.eval()
        self.rnn_model = rnn_model
        self.rnn_optimizer = rnn_optimizer
        self.rnn_agent = AgentNetwork(rnn_model, self.n_actions, deterministic=True, device=self.device)
        
        if self.verbose:
            print('\nRNN training finished.')
            print(f'Training took {time.time() - start_time:.2f} seconds.')

        if self.save_path_rnn is not None:
            print(f'Saving RNN model to {self.save_path_rnn}...')
            self.save_spice(self.save_path_rnn, None)
            
        # ------------------------------------------------------------------------
        # Fit SPICE
        # ------------------------------------------------------------------------
        
        self.spice_agent = {}
        self.spice_features = {}
        spice_modules = {rnn_module: {} for rnn_module in self.rnn_modules}

        self.spice_agent = fit_spice(
            rnn_modules=self.rnn_modules,
            control_signals=self.control_parameters,
            agent_rnn=self.rnn_agent,
            data=dataset,
            polynomial_degree=self.spice_library_polynomial_degree,
            library_setup=self.spice_library_config,
            filter_setup=self.spice_filter_config,
            optimizer_type=self.spice_optimizer_type,
            optimizer_threshold=self.spice_optim_threshold,
            optimizer_alpha=self.spice_optim_regularization,
            shuffle = False,
            n_trials_off_policy = self.n_trials_off_policy,
            n_sessions_off_policy = self.n_sessions_off_policy,
            n_trials_same_action_off_policy = self.n_trials_same_action_off_policy,
            train_test_ratio = self.train_test_ratio,
            deterministic = self.deterministic,
            verbose = self.verbose,
            use_optuna = self.use_optuna,
            optuna_threshold = self.optuna_threshold,
            optuna_n_trials = self.optuna_n_trials,
        )[0]

        # self.spice_features = self.spice_agent.get_spice_features()
        
        if self.verbose:
            print('SPICE training finished.')
            print(f'Training took {time.time() - start_time:.2f} seconds.')

        if self.save_path_spice is not None:
            print(f'Saving SPICE model to {self.save_path_spice}...')
            self.save_spice(None, self.save_path_spice)

    def predict(self, conditions: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Make predictions using both RNN and SPICE models.
        
        Args:
            conditions: Array of shape (n_participants, n_trials, n_features)
            
        Returns:
            Tuple containing:
            - RNN predictions
            - SPICE predictions
        """
        
        if isinstance(conditions, np.ndarray):
            conditions = torch.tensor(conditions, dtype=torch.float32, device=self.device)
        elif isinstance(conditions, torch.Tensor):
            pass
        else:
            raise TypeError(f"conditions must be either of type numpy.ndarray or torch.Tensor.")
        
        # get predictions about action probability by RNN and SPICE separately
        prediction_rnn = np.full((*conditions.shape[:-1], self.n_actions), np.nan).reshape(-1, self.n_actions)
        prediction_spice = np.full((*conditions.shape[:-1], self.n_actions), np.nan).reshape(-1, self.n_actions)
        mask = torch.sum(conditions[..., :self.n_actions].reshape(-1, self.n_actions), dim=-1, keepdim=False) != -2
        
        # rnn predictions
        prediction = self.rnn_agent._model(conditions, batch_first=True)[0].reshape(-1, self.n_actions)
        prediction = torch.nn.functional.softmax(prediction, dim=-1).detach().cpu().numpy()
        prediction_rnn[mask.detach().cpu().numpy()] = prediction[mask]
        prediction_rnn = prediction_rnn.reshape(*conditions.shape[:-1], self.n_actions)
        
        # SPICE predictions
        prediction = self.spice_agent._model(conditions, batch_first=True)[0].reshape(-1, self.n_actions)
        prediction = torch.nn.functional.softmax(prediction, dim=-1).detach().cpu().numpy()
        prediction_spice[mask.detach().cpu().numpy()] = prediction[mask]
        prediction_spice = prediction_spice.reshape(*conditions.shape[:-1], self.n_actions)
        
        return prediction_rnn, prediction_spice

    def print_spice_model(self, participant_id: int = 0) -> None:
        """
        Get the learned SPICE features and equations.
        
        Returns:
            Dictionary containing features and equations for each agent/model
        """
        print("SPICE modules:")
        self.spice_agent.print_model(participant_id)
        
        if hasattr(self.rnn_model, 'betas') and len(self.rnn_model.betas) > 0:
            for key in self.rnn_model.betas:
                if not isinstance(self.rnn_model.betas[key], ParameterModule):
                    participant_embedding = self.rnn_model.participant_embedding(torch.tensor(participant_id, device=self.device).int())
                else:
                    participant_embedding = None
                print(f"beta({key}) = {self.rnn_model.betas[key](participant_embedding).item():.4f}")
    
    def get_spice_agents(self) -> Dict:
        """
        Get the trained SPICE agents.
        
        Returns:
            Dictionary containing SPICE agents
        """
        if self.spice_agent is None:
            raise ValueError("Model hasn't been fitted yet. Call fit() first.")
        return self.spice_agent

    def get_participant_embeddings(self) -> Dict:
        if hasattr(self.rnn_model, 'participant_embedding'):
            participant_ids = torch.arange(self.n_participants, device=self.device, dtype=torch.int32).view(-1, 1)
            embeddings = self.rnn_model.participant_embedding(participant_ids)
            return {participant_id.item(): embeddings[participant_id, 0] for participant_id in participant_ids}
        else:
            print(f'RNN model has no participant_embedding module.')
            return None
        
    def load_rnn_model(self, path_model: str, deterministic: bool = True):
        self.rnn_model, self.rnn_optimizer = load_checkpoint(path_model, self.rnn_model, self.optimizer_rnn)
        self.rnn_agent = AgentNetwork(self.rnn_model, self.n_actions, deterministic=deterministic, device=self.device)
        
    def load_spice_model(self, path_spice: str, deterministic: bool = True):
        spice_modules = load_spice(path_spice)
        self.spice_agent = AgentSpice(model_rnn=self.rnn_agent._model, sindy_modules=spice_modules, n_actions=self.rnn_agent._n_actions, deterministic=deterministic)

    def load_spice(self, path_rnn: str, path_spice: str, deterministic: bool = True):
        """
        Load the RNN and SPICE models from the given paths.
        
        Args:
            path_rnn: Path to the RNN model
            path_spice: Path to the SPICE model
            deterministic: Whether to use a deterministic model (default: True)
        """
        if path_rnn is not None:
            self.load_rnn_model(path_rnn, deterministic=deterministic)
        if path_spice is not None:
            self.load_spice_model(path_spice, deterministic=deterministic)

    def save_spice(self, path_rnn: str = None, path_spice: str = None):
        """
        Save the RNN and SPICE models to the given paths.
        If path_rnn is None, only the SPICE model will be saved (requires a fitted SPICE model including the RNN).
        If path_spice is None, only the RNN model will be saved (requires a fitted RNN model).
        
        Args:
            path_rnn: Path to the RNN model
            path_spice: Path to the SPICE model
        """
        if path_rnn is not None:
            # Save RNN model
            state_dict = {'model': self.rnn_model.state_dict(), 'optimizer': self.rnn_optimizer.state_dict()}
            torch.save(state_dict, path_rnn)
        if path_spice is not None:
            # Save SPICE model
            save_spice(self.spice_agent, path_spice)


import torch
from torch.utils.data import Dataset
from typing import Union, List


class DatasetRNN(Dataset):
    def __init__(
        self, 
        xs: torch.Tensor, 
        ys: torch.Tensor,
        normalize_features: tuple[int] = None, 
        sequence_length: int = None,
        stride: int = 1,
        device=None,
        ):
        """Initializes the dataset for training the RNN. Holds information about the previous actions and rewards as well as the next action.
        Actions can be either one-hot encoded or indexes.

        Args:
            xs (torch.Tensor): Actions and rewards in the shape (n_sessions, n_timesteps, n_features)
            ys (torch.Tensor): Next action
            batch_size (Optional[int], optional): Sets batch size if desired else uses n_samples as batch size.
            device (torch.Device, optional): Torch device. If None, uses cuda if available else cpu.
        """
        
        if device is None:
            device = torch.device('cpu')
        
        # check for type of xs and ys
        if not isinstance(xs, torch.Tensor):
            xs = torch.tensor(xs, dtype=torch.float32)
        if not isinstance(ys, torch.Tensor):
            ys = torch.tensor(ys, dtype=torch.float32)
        
        # check dimensions of xs and ys
        if len(xs.shape) == 2:
            xs = xs.unsqueeze(0)
        if len(ys.shape) == 2:
            ys = ys.unsqueeze(0)
            
        if normalize_features is not None:
            if isinstance(normalize_features, int):
                normalize_features = tuple(normalize_features)
            for feature in normalize_features:
                xs[:, :, feature] = self.normalize(xs[:, :, feature])
        
        # normalize data
        # x_std = xs.std(dim=(0, 1))
        # x_mean = xs.mean(dim=(0, 1))
        # xs = (xs - x_mean) / x_std
        # ys = (ys - x_mean) / x_std
        
        sequence_length = None if sequence_length == -1 else sequence_length
        self.sequence_length = sequence_length if sequence_length is not None else xs.shape[1]
        self.stride = stride
        
        if sequence_length is not None:
            xs, ys = self.set_sequences(xs, ys)
        self.device = device
        self.xs = xs.to(device)
        self.ys = ys.to(device)
        
    def normalize(self, data):
        x_min = torch.min(data)
        x_max = torch.max(data)
        return (data - x_min) / (x_max - x_min)
        
    def set_sequences(self, xs, ys):
        # sets sequences of length sequence_length with specified stride from the dataset
        xs_sequences = []
        ys_sequences = []
        for i in range(0, max(1, xs.shape[1]-self.sequence_length), self.stride):
            xs_sequences.append(xs[:, i:i+self.sequence_length, :])
            ys_sequences.append(ys[:, i:i+self.sequence_length, :])
        xs = torch.cat(xs_sequences, dim=0)
        ys = torch.cat(ys_sequences, dim=0)
        
        if len(xs.shape) == 2:
            xs = xs.unsqueeze(1)
            ys = ys.unsqueeze(1)
            
        return xs, ys
    
    def __len__(self):
        return self.xs.shape[0]
    
    def __getitem__(self, idx):
        return self.xs[idx, :], self.ys[idx, :]


def load_checkpoint(params_path, model, optimizer):
    # load trained parameters
    state_dict = torch.load(params_path, map_location=torch.device('cpu'))
    
    state_dict_model = state_dict['model']
    state_dict_optimizer = state_dict['optimizer']
    
    model.load_state_dict(state_dict_model)
    optimizer.load_state_dict(state_dict_optimizer)
    
    return model, optimizer


def parameter_file_naming(params_path, alpha_reward, alpha_penalty, alpha_counterfactual, confirmation_bias, forget_rate, beta_reward, alpha_choice, beta_choice, variance, verbose=False):
    # create name for corresponding rnn
  
    params_path += '_rnn'
    
    params_path += '_br' + str(beta_reward).replace('.', '')
    
    if alpha_reward > 0:
        params_path += '_a' + str(alpha_reward).replace('.', '')
    
    if alpha_penalty >= 0:
        params_path += '_ap' + str(alpha_penalty).replace('.', '')
    
    if alpha_counterfactual > 0:
        params_path += '_ac' + str(alpha_counterfactual).replace('.', '')
        
    if beta_choice > 0 and alpha_choice > 0:
        params_path += '_bch' + str(beta_choice).replace('.', '')
        params_path += '_ach' + str(alpha_choice).replace('.', '')
    
    if forget_rate > 0:
        params_path += '_f' + str(forget_rate).replace('.', '')
        
    if confirmation_bias > 0:
        params_path += '_cb' + str(confirmation_bias).replace('.', '')
    
    if not isinstance(variance, dict):
        if variance != 0:
            params_path += '_var' + str(variance).replace('.', '').replace('-1','Mean')
    else:
        params_path += '_varDict'
        
    # if non_binary_reward:
    #     params_path += '_nonbinary'
        
    params_path += '.pkl'
    
    if verbose:
        print(f'Automatically generated name for model parameter file: {params_path}.')
        
    return params_path


def split_data_along_timedim(dataset: DatasetRNN, split_ratio: float, device: torch.device = torch.device('cpu')):
    """Split the data along the time dimension (dim=1). 
    Each session (dim=0) can be of individual length and is therefore post-padded with -1.
    To split the data into training and testing samples according to the split_ratio each session's individual length has to be considered.
    E.g.: 
    1. session_length = len(session 0) -> 120
    2. split_index = int(split_ratio * session length) -> 96
    3. samples for training data -> 96
    4. samples for testing data -> 24 = 120 - 96    

    Args:
        data (torch.Tensor): Data containing all sessions in the shape (session, time, features)
        split_ratio (float): Float number indicating the ratio to be used as training data

    Returns:
        tuple(DatasetRNN, DatasetRNN): Training data and testing data splitted along time dimension (dim=1)
    """
    
    dim = 1
    xs, ys = dataset.xs, dataset.ys
    
    # Create a mask of non-zero elements
    non_zero_mask = (xs[..., 0] != -1).int()
    
    # Find cumulative sum along the specified dimension in reverse order
    cumsum = torch.cumsum(non_zero_mask, dim)
    
    # Find the index where the cumulative sum becomes 1 in the reversed array
    last_nonzero_indices = torch.argmax(cumsum, dim=dim)
    
    # compute the indeces at which the data is going to be splitted into training and testing data
    split_indices = (last_nonzero_indices * split_ratio).int()
    
    # initialize training data and testing data storages
    train_xs = torch.zeros((xs.shape[0], max(split_indices), xs.shape[2]), device=dataset.device) - 1
    test_xs = torch.zeros((xs.shape[0], max(last_nonzero_indices - split_indices), xs.shape[2]), device=dataset.device) - 1
    train_ys = torch.zeros((xs.shape[0], max(split_indices), ys.shape[2]), device=dataset.device) - 1
    test_ys = torch.zeros((xs.shape[0], max(last_nonzero_indices - split_indices), ys.shape[2]), device=dataset.device) - 1
    
    # get columns which had no -1 values in the first place to fill them up entirely in the training and testing data
    # necessary for e.g. participant-IDs because otherwise -1 will be passed to embedding layer -> Error
    example_session_id = torch.argmax((last_nonzero_indices < xs.shape[1]-1).int()).item()
    full_columns = xs[example_session_id, -1] != -1
    
    # fill up training and testing data
    for index_session in range(xs.shape[0]):
        train_xs[index_session, :split_indices[index_session]] = xs[index_session, :split_indices[index_session]]
        test_xs[index_session, :last_nonzero_indices[index_session]-split_indices[index_session]] = xs[index_session, split_indices[index_session]:last_nonzero_indices[index_session]]
        train_ys[index_session, :split_indices[index_session]] = ys[index_session, :split_indices[index_session]]
        test_ys[index_session, :last_nonzero_indices[index_session]-split_indices[index_session]] = ys[index_session, split_indices[index_session]:last_nonzero_indices[index_session]]

        # fill up non-"-1"-columns (only applicable for xs)
        train_xs[index_session, :, full_columns] = xs[index_session, :train_xs.shape[1], full_columns]
        test_xs[index_session, :, full_columns] = xs[index_session, :test_xs.shape[1], full_columns]
    
    return DatasetRNN(train_xs, train_ys, device=device), DatasetRNN(test_xs, test_ys, device=device)


def split_data_along_sessiondim(dataset: DatasetRNN, list_test_sessions: List[int] = None, device: torch.device = torch.device('cpu')):
    """Split the data along the time dimension (dim=1). 
    Each session (dim=0) can be of individual length and is therefore post-padded with -1.
    To split the data into training and testing samples according to the split_ratio each session's individual length has to be considered.
    E.g.: 
    1. session_length = len(session 0) -> 120
    2. split_index = int(split_ratio * session length) -> 96
    3. samples for training data -> 96
    4. samples for testing data -> 24 = 120 - 96    

    Args:
        data (torch.Tensor): Data containing all sessions in the shape (session, time, features)
        split_ratio (float): Float number indicating the ratio to be used as training data

    Returns:
        tuple(DatasetRNN, DatasetRNN): Training data and testing data splitted along time dimension (dim=1)
    """
    
    if list_test_sessions is not None:
        
        dim = 1
        xs, ys = dataset.xs.cpu(), dataset.ys.cpu()
        
        # get participant ids
        participants_ids = xs[:, 0, -1].unique()
        
        # get sessions ids
        session_ids = xs[:, 0, -3].unique()
        
        # set training sessions
        if list_test_sessions:
            n_sessions_test = len(list_test_sessions)
            session_ids_test = torch.tensor(list_test_sessions, dtype=torch.float32)
        else:
            n_sessions_test = 0
            session_ids_test = torch.tensor()
        
        n_sessions_train = len(session_ids) - n_sessions_test
            
        if all(session_ids[:-1] < 1):
            session_ids_test /= len(session_ids) - 1
        
        # set test sessions
        session_ids_train = torch.zeros(n_sessions_train)
        index_sid = 0
        for sid in session_ids:
            if not sid in session_ids_test:
                session_ids_train[index_sid] = sid
                index_sid += 1     
        
        # setup new variables
        train_xs, test_xs, train_ys, test_ys = torch.zeros((len(participants_ids) * n_sessions_train, *xs.shape[1:])), torch.zeros((len(participants_ids) * n_sessions_test, *xs.shape[1:])), torch.zeros((len(participants_ids) * n_sessions_train, *ys.shape[1:]))-1, torch.zeros((len(participants_ids) * n_sessions_test, *ys.shape[1:]))-1
        train_xs[..., :-1], test_xs[..., :-1] = train_xs[..., :-1]-1, test_xs[..., :-1]-1
        
        index_train = 0
        index_test = 0
        for pid in participants_ids:
            for sid in session_ids:
                mask_ids = torch.logical_and(xs[:, 0, -3] == sid, xs[:, 0, -1] == pid)
                if sid in session_ids_train:
                    train_xs[index_train] = xs[mask_ids]
                    train_ys[index_train] = ys[mask_ids]
                    index_train += 1
                elif sid in session_ids_test:
                    test_xs[index_test] = xs[mask_ids]
                    test_ys[index_test] = ys[mask_ids]
                    index_test += 1
                else:
                    raise ValueError("session id was not found in training nor test sessions.")
        
        
        return DatasetRNN(train_xs, train_ys, device=device), DatasetRNN(test_xs, test_ys, device=device)

    else:
        return dataset, dataset

def reshape_data_along_participantdim(dataset: DatasetRNN, device: torch.device = torch.device('cpu')):
    """Reshape the data along the participant dim.
    
    Args:
        dataset (DatasetRNN): current dataset of shape (participants*sessions*, trial, features)
        device (torch.device, optional): Defaults to torch.device('cpu').

    Returns:
        DatasetRNN: restructured dataset with shape (participant, session, trial, features)
    """
    
    xs, ys = dataset.xs.cpu(), dataset.ys.cpu()
    
    # get participant ids
    participants_ids = xs[:, 0, -1].unique()
    
    # setup new variables
    xs_new, ys_new = [], []
    n_sessions_max = 0
    for pid in participants_ids:
        # collect participant-level sessions
        mask_ids = xs[:, 0, -1] == pid
        xs_new.append(xs[mask_ids])
        ys_new.append(ys[mask_ids])
        # correct number of sessions for each participant if maximum number of session changed
        if len(xs[mask_ids]) > n_sessions_max:
            n_sessions_max = max(len(xs[mask_ids]), n_sessions_max)
            for xs_i, ys_i in zip(xs_new, ys_new):
                if xs_i.shape[0] < n_sessions_max:
                    pad_sessions_xs = torch.zeros((n_sessions_max-xs_i.shape[0], *xs_i.shape[1:]), device=xs_i.device)
                    # set ID dimensions as full arrays without -1
                    pad_sessions_xs[..., :-3] += -1
                    # setting only participant ID makes sense because experiment and session IDs do not affect the following computations
                    pad_sessions_xs[..., -1] += xs_i[0, 0, -1]
                    xs_i = torch.concat((xs_i, pad_sessions_xs))
                    pad_sessions_ys = torch.zeros((n_sessions_max-ys_i.shape[0], *ys_i.shape[1:]), device=ys_i.device) - 1
                    ys_i = torch.concat((ys_i, pad_sessions_ys))
    
    # setup new dataset with shape (participant, session, trial, features)
    return DatasetRNN(torch.stack(xs_new), torch.stack(ys_new), device=device)


from setuptools import setup
import os

# Read requirements
with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = [line.strip() for line in fh if line.strip() and not line.startswith("#")]

# Read dev requirements if file exists
dev_requirements = []
if os.path.exists("requirements-dev.txt"):
    with open("requirements-dev.txt", "r", encoding="utf-8") as fh:
        dev_requirements = [line.strip() for line in fh if line.strip() and not line.startswith("#")]
else:
    dev_requirements = [
        "pytest>=7.0.0",
        "pytest-cov>=4.0.0",
        "black>=23.0.0",
        "flake8>=6.0.0",
        "mypy>=1.0.0",
        "isort>=5.0.0",
    ]

setup(
    install_requires=requirements,
    extras_require={
        "dev": dev_requirements,
    },
) 


from spice.resources.rnn import BaseRNN
from spice.estimator import SpiceConfig
import torch


RESCOLA_WAGNER_CONFIG = SpiceConfig(
    library_setup={
        'x_value_reward_chosen': ['c_reward'],
    },
    filter_setup={
        'x_value_reward_chosen': ['c_action', 1, True],
    },
    control_parameters=['c_action', 'c_reward'],
    rnn_modules=['x_value_reward_chosen']
)


class RescorlaWagnerRNN(BaseRNN):
    init_values={
        'x_value_reward': 0.5,
    }

    def __init__(
        self,
        n_actions,
        **kwargs,
    ):   
        super(RescorlaWagnerRNN, self).__init__(n_actions=n_actions)
        
        # set up the submodules
        self.submodules_rnn['x_value_reward_chosen'] = self.setup_module(input_size=1)
        
    def forward(self, inputs, prev_state=None, batch_first=False):
        """Forward pass of the RNN

        Args:
            inputs (torch.Tensor): includes all necessary inputs (action, reward, participant id) to the RNN to let it compute the next action
            prev_state (Tuple[torch.Tensor], optional): That's the previous memory state of the RNN containing the reward-based value. Defaults to None.
            batch_first (bool, optional): Indicates whether the first dimension of inputs is batch (True) or timesteps (False). Defaults to False.
        """
        
        # First, we have to initialize all the inputs and outputs (i.e. logits)
        inputs, ids, logits, timesteps = self.init_forward_pass(inputs, prev_state, batch_first)
        actions, rewards, _, _ = inputs
        
        for timestep, action, reward in zip(timesteps, actions, rewards):
            
            # record the inputs for training SINDy later on
            self.record_signal('c_action', action)
            self.record_signal('c_reward', reward)
            self.record_signal('x_value_reward_chosen', self.state['x_value_reward'])
            
            # Let's perform the belief update for the reward-based value of the chosen option
            # since all values are given to the rnn-module (independent of each other), the chosen value is selected by setting the action to the chosen one
            # if we would like to perform a similar update by calling a rnn-module for the non-chosen action, we would set the parameter to action=1-action.
            next_value_reward_chosen = self.call_module(
                key_module='x_value_reward_chosen',
                key_state='x_value_reward',
                action=action,
                inputs=reward,
                )

            # and keep the value of the not-chosen option unchanged
            next_value_reward_not_chosen = self.state['x_value_reward'] * (1-action)
            
            self.state['x_value_reward'] = next_value_reward_chosen + next_value_reward_not_chosen  # memory state = (0.8, 0.3) <- next_value = (0.8, 0) + (0, 0.3)
            
            # Now keep track of this value in the output array
            logits[timestep] = self.state['x_value_reward']
        
        # post-process the forward pass; give here as inputs the logits, batch_first and all values from the memory state
        # self.state['x_value_reward'] = value_reward
        logits = self.post_forward_pass(logits, batch_first)
        
        return logits, self.get_state()
    

FORGETTING_RNN_CONFIG = SpiceConfig(
    # Add already here the new module and update the library and filter setup.
    rnn_modules=['x_value_reward_chosen', 'x_value_reward_not_chosen'],
    
    control_parameters=['c_action', 'c_reward'],

    # The new module which handles the not-chosen value, does not need any additional inputs except for the value
    library_setup={
        'x_value_reward_chosen': ['c_reward'],
        'x_value_reward_not_chosen': [],
    },

    # Further, the new module should be applied only to the not-chosen values
    filter_setup={
        'x_value_reward_chosen': ['c_action', 1, True],
        'x_value_reward_not_chosen': ['c_action', 0, True],
    },
)

class ForgettingRNN(BaseRNN):
    
    init_values = {
        'x_value_reward': 0.5,
    }
    
    def __init__(
        self,
        n_actions,
        **kwargs,
    ):
        super(ForgettingRNN, self).__init__(n_actions=n_actions)
        
        # set up the submodules
        self.submodules_rnn['x_value_reward_chosen'] = self.setup_module(input_size=1)
        self.submodules_rnn['x_value_reward_not_chosen'] = self.setup_module(input_size=0)
        
    def forward(self, inputs, prev_state=None, batch_first=False):
        """Forward pass of the RNN

        Args:
            inputs (torch.Tensor): includes all necessary inputs (action, reward, participant id) to the RNN to let it compute the next action
            prev_state (Tuple[torch.Tensor], optional): That's the previous memory state of the RNN containing the reward-based value. Defaults to None.
            batch_first (bool, optional): Indicates whether the first dimension of inputs is batch (True) or timesteps (False). Defaults to False.
        """
        
        # First, we have to initialize all the inputs and outputs (i.e. logits)
        inputs, ids, logits, timesteps = self.init_forward_pass(inputs, prev_state, batch_first)
        actions, rewards, _, _ = inputs
                
        for timestep, action, reward in zip(timesteps, actions, rewards):
            
            # record the inputs for training SINDy later on
            self.record_signal('c_action', action)
            self.record_signal('c_reward', reward)
            self.record_signal('x_value_reward_chosen', self.state['x_value_reward'])
            self.record_signal('x_value_reward_not_chosen', self.state['x_value_reward'])
            
            # Let's perform the belief update for the reward-based value of the chosen option
            next_value_reward_chosen = self.call_module(
                key_module='x_value_reward_chosen',
                key_state='x_value_reward',
                action=action,
                inputs=reward,
                )

            # Now a RNN-module updates the not-chosen reward-based value instead of keeping it the same
            next_value_reward_not_chosen = self.call_module(
                key_module='x_value_reward_not_chosen',
                key_state='x_value_reward',
                action=1-action,
                inputs=None,
                )
            
            # keep track of the updated value in the memory state
            self.state['x_value_reward'] = next_value_reward_chosen + next_value_reward_not_chosen
            
            # Now keep track of this value in the output array
            logits[timestep] = self.state['x_value_reward']
        
        # post-process the forward pass; give here as inputs the logits, batch_first and all values from the memory state
        logits = self.post_forward_pass(logits, batch_first)
        
        return logits, self.get_state()
    

LEARNING_RATE_RNN_CONFIG = SpiceConfig(
    rnn_modules=['x_learning_rate_reward', 'x_value_reward_not_chosen'],
    
    control_parameters=['c_action', 'c_reward', 'c_value_reward'],

    library_setup={
        'x_learning_rate_reward': ['c_reward', 'c_value_reward'],
        'x_value_reward_not_chosen': [],
    },

    filter_setup={
        'x_learning_rate_reward': ['c_action', 1, True],
        'x_value_reward_not_chosen': ['c_action', 0, True],
    },
)


class LearningRateRNN(BaseRNN):
    
    init_values = {
            'x_value_reward': 0.5,
            'x_learning_rate_reward': 0,
    }
    
    def __init__(
        self,
        n_actions,
        **kwargs,
    ):
        super(LearningRateRNN, self).__init__(n_actions=n_actions)
        
        # set up the submodules
        self.submodules_rnn['x_learning_rate_reward'] = self.setup_module(input_size=2)
        self.submodules_rnn['x_value_reward_not_chosen'] = self.setup_module(input_size=0)

        # set up hard-coded equations
        # add here a RNN-module in the form of an hard-coded equation to compute the update for the chosen reward-based value
        self.submodules_eq['x_value_reward_chosen'] = lambda value, inputs: value + inputs[..., 1] * (inputs[..., 0] - value)

        # add a scaling factor (i.e. inverse noise temperature) for 'x_value_reward'
        # self.betas = torch.nn.ParameterDict()
        # self.betas['x_value_reward'] = torch.nn.Parameter(torch.tensor(1.))             
        self.betas['x_value_reward'] = self.setup_constant()
        
    def forward(self, inputs, prev_state=None, batch_first=False):
        """Forward pass of the RNN

        Args:
            inputs (torch.Tensor): includes all necessary inputs (action, reward, participant id) to the RNN to let it compute the next action
            prev_state (Tuple[torch.Tensor], optional): That's the previous memory state of the RNN containing the reward-based value. Defaults to None.
            batch_first (bool, optional): Indicates whether the first dimension of inputs is batch (True) or timesteps (False). Defaults to False.
        """
        
        # First, we have to initialize all the inputs and outputs (i.e. logits)
        inputs, ids, logits, timesteps = self.init_forward_pass(inputs, prev_state, batch_first)
        actions, rewards, _, _ = inputs
                
        for timestep, action, reward in zip(timesteps, actions, rewards):
            
            # record the inputs for training SINDy later on
            self.record_signal('c_action', action)
            self.record_signal('c_reward', reward)
            self.record_signal('c_value_reward', self.state['x_value_reward'])
            self.record_signal('x_value_reward_not_chosen', self.state['x_value_reward'])
            self.record_signal('x_learning_rate_reward', self.state['x_learning_rate_reward'])
            
            # Let's compute the learning rate dynamically
            # Now we have to use a sigmoid activation function on the output learning rate to constrain it to a value range of (0, 1)
            # this is necessary for two reasons:
            #   1. Preventing exploding gradients
            #   2. Remember the found equation for 'x_value_reward_chosen' from before: 
            #       The learning rate was scaled according to the magnitudes of the reward and the actual value 
            #       e.g. for the reward: alpha*beta -> alpha * beta = 0.3 * 3 = 0.9 and for the reward-based value: 1-alpha = 1 - 0.3 = 0.7
            #       The hard-coded equation for the reward-prediction error does not permit this flexibility. 
            #       But we can circumvein this by applying the sigmoid activation to the learning rate to staying conform with the reward-prediction error
            #       and later applying the inverse noise temperature (i.e. trainable parameter) to the updated value 
            learning_rate_reward = self.call_module(
                key_module='x_learning_rate_reward',
                key_state='x_learning_rate_reward',
                action=action,
                inputs=(reward, self.state['x_value_reward']),
                activation_rnn=torch.nn.functional.sigmoid,
            )
            
            # Let's perform the belief update for the reward-based value of the chosen option            
            next_value_reward_chosen = self.call_module(
                key_module='x_value_reward_chosen',
                key_state='x_value_reward',
                action=action,
                inputs=(reward, learning_rate_reward),
                )
            
            # Update of the not-chosen reward-based value
            next_value_reward_not_chosen = self.call_module(
                key_module='x_value_reward_not_chosen',
                key_state='x_value_reward',
                action=1-action,
                inputs=None,
                )
            
            # updating the memory state
            self.state['x_learning_rate_reward'] = learning_rate_reward
            self.state['x_value_reward'] = next_value_reward_chosen + next_value_reward_not_chosen
            
            # Now keep track of the logit in the output array
            logits[timestep] = self.state['x_value_reward'] * self.betas['x_value_reward']()
        
        # post-process the forward pass; give here as inputs the logits, batch_first and all values from the memory state
        logits = self.post_forward_pass(logits, batch_first)
        
        return logits, self.get_state()
    

# The participant embedding RNN is basically the learning rate RNN with an additional participant embedding layer
PARTICIPANT_EMBEDDING_RNN_CONFIG = LEARNING_RATE_RNN_CONFIG


class ParticipantEmbeddingRNN(BaseRNN):
    
    init_values = {
            'x_value_reward': 0.5,
            'x_learning_rate_reward': 0,
        }
    
    def __init__(
        self,
        n_actions,
        # add an additional inputs to set the number of participants in your data
        n_participants,
        embedding_size: int = 8,
        **kwargs,
    ):
        
        super(ParticipantEmbeddingRNN, self).__init__(n_actions=n_actions, embedding_size=embedding_size)
        
        # specify here the participant embedding
        self.participant_embedding = self.setup_embedding(n_participants, embedding_size, dropout=0.)
        
        # Add a scaling factor (inverse noise temperature) for each participant.
        self.betas['x_value_reward'] = self.setup_constant(self.embedding_size)
        
        # and here we specify the general architecture
        # add to the input_size the embedding_size as well because we are going to pass the participant-embedding to the RNN-modules
        # set up the submodules
        self.submodules_rnn['x_learning_rate_reward'] = self.setup_module(input_size=2+self.embedding_size)
        self.submodules_rnn['x_value_reward_not_chosen'] = self.setup_module(input_size=0+self.embedding_size)
        
        # set up hard-coded equations
        # add here a RNN-module in the form of an hard-coded equation to compute the update for the chosen reward-based value
        self.submodules_eq['x_value_reward_chosen'] = lambda value, inputs: value + inputs[..., 1] * (inputs[..., 0] - value)
        
    def forward(self, inputs, prev_state=None, batch_first=False):
        """Forward pass of the RNN

        Args:
            inputs (torch.Tensor): includes all necessary inputs (action, reward, participant id) to the RNN to let it compute the next action
            prev_state (Tuple[torch.Tensor], optional): That's the previous memory state of the RNN containing the reward-based value. Defaults to None.
            batch_first (bool, optional): Indicates whether the first dimension of inputs is batch (True) or timesteps (False). Defaults to False.
        """
        
        # First, we have to initialize all the inputs and outputs (i.e. logits)
        inputs, ids, logits, timesteps = self.init_forward_pass(inputs, prev_state, batch_first)
        actions, rewards, _, _ = inputs
        participant_id, _ = ids

        # Here we compute now the participant embeddings for each entry in the batch
        participant_embedding = self.participant_embedding(participant_id[:, 0].int())
        
        for timestep, action, reward in zip(timesteps, actions, rewards):
            
            # record the inputs for training SINDy later on
            self.record_signal('c_action', action)
            self.record_signal('c_reward', reward)
            self.record_signal('c_value_reward', self.state['x_value_reward'])
            self.record_signal('x_learning_rate_reward', self.state['x_learning_rate_reward'])
            self.record_signal('x_value_reward_not_chosen', self.state['x_value_reward'])
            
            learning_rate_reward = self.call_module(
                key_module='x_learning_rate_reward',
                key_state='x_learning_rate_reward',
                action=action,
                inputs=(reward, self.state['x_value_reward']),
                # add participant-embedding (for RNN-modules) and participant-index (later for SINDy-modules) 
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
            )
            
            # Let's perform the belief update for the reward-based value of the chosen option            
            next_value_reward_chosen = self.call_module(
                key_module='x_value_reward_chosen',
                key_state='x_value_reward',
                action=action,
                inputs=(reward, learning_rate_reward),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                )
            
            # Update of the not-chosen reward-based value
            next_value_reward_not_chosen = self.call_module(
                key_module='x_value_reward_not_chosen',
                key_state='x_value_reward',
                action=1-action,
                inputs=None,
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                )
            
            self.state['x_value_reward'] = next_value_reward_chosen + next_value_reward_not_chosen
            self.state['x_learning_rate_reward'] = learning_rate_reward
            
            # Now keep track of the logit in the output array
            logits[timestep] = self.state['x_value_reward'] * self.betas['x_value_reward'](participant_embedding)
        
        # post-process the forward pass; give here as inputs the logits, batch_first and all values from the memory state
        logits = self.post_forward_pass(logits, batch_first)
        
        return logits, self.get_state()
    

WEINHARDT_2024_CONFIG = SpiceConfig(
    rnn_modules=['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen'],
    control_parameters=['c_action', 'c_reward', 'c_value_reward'],
    # The new module which handles the not-chosen value, does not need any additional inputs except for the value
    library_setup = {
        # 'x_value_reward_chosen': ['c_reward'] -> Remove this one from the library as we are not going to identify the dynamics of a hard-coded equation
        'x_learning_rate_reward': ['c_reward', 'c_value_reward'],
        'x_value_reward_not_chosen': [],
        'x_value_choice_chosen': [],
        'x_value_choice_not_chosen': [],
    },

    # Further, the new module should be applied only to the not-chosen values
    filter_setup = {
        # 'x_value_reward_chosen': ['c_action', 1, True], -> Remove this one as well
        'x_learning_rate_reward': ['c_action', 1, True],
        'x_value_reward_not_chosen': ['c_action', 0, True],
        'x_value_choice_chosen': ['c_action', 1, True],
        'x_value_choice_not_chosen': ['c_action', 0, True],
    }    
)


class Weinhardt2024RNN(BaseRNN):
    
    init_values = {
            'x_value_reward': 0.5,
            'x_value_choice': 0.,
            'x_learning_rate_reward': 0.,
        }
    
    def __init__(
        self,
        n_actions,
        n_participants,
        dropout=0.5,
        **kwargs,
    ):
        
        super(Weinhardt2024RNN, self).__init__(n_actions=n_actions)
        
        # set up the participant-embedding layer
        self.embedding_size = 8
        self.participant_embedding = torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=n_participants, embedding_dim=self.embedding_size),
            torch.nn.ReLU(),
            torch.nn.Dropout(p=dropout),
            )
        
        # scaling factor (inverse noise temperature) for each participant for the values which are handled by an hard-coded equation
        self.betas['x_value_reward'] = torch.nn.Sequential(torch.nn.Linear(self.embedding_size, 1), torch.nn.ReLU())
        self.betas['x_value_choice'] = torch.nn.Sequential(torch.nn.Linear(self.embedding_size, 1), torch.nn.ReLU())
        
        # set up the submodules
        self.submodules_rnn['x_learning_rate_reward'] = self.setup_module(input_size=2+self.embedding_size)
        self.submodules_rnn['x_value_reward_not_chosen'] = self.setup_module(input_size=0+self.embedding_size)
        self.submodules_rnn['x_value_choice_chosen'] = self.setup_module(input_size=0+self.embedding_size)
        self.submodules_rnn['x_value_choice_not_chosen'] = self.setup_module(input_size=0+self.embedding_size)
        
        # set up hard-coded equations
        self.submodules_eq['x_value_reward_chosen'] = lambda value, inputs: value + inputs[..., 1] * (inputs[..., 0] - value)
        
    def forward(self, inputs, prev_state=None, batch_first=False):
        """Forward pass of the RNN

        Args:
            inputs (torch.Tensor): includes all necessary inputs (action, reward, participant id) to the RNN to let it compute the next action
            prev_state (Tuple[torch.Tensor], optional): That's the previous memory state of the RNN containing the reward-based value. Defaults to None.
            batch_first (bool, optional): Indicates whether the first dimension of inputs is batch (True) or timesteps (False). Defaults to False.
        """
        
        # First, we have to initialize all the inputs and outputs (i.e. logits)
        inputs, embedding_variables, logits, timesteps = self.init_forward_pass(inputs, prev_state, batch_first)
        actions, rewards, _, _ = inputs
        participant_id, _ = embedding_variables
        
        # Here we compute now the participant embeddings for each entry in the batch
        participant_embedding = self.participant_embedding(participant_id[:, 0].int())
        beta_reward = self.betas['x_value_reward'](participant_embedding)
        beta_choice = self.betas['x_value_choice'](participant_embedding)
        
        for timestep, action, reward in zip(timesteps, actions, rewards):
            
            # record the inputs for training SINDy later on
            self.record_signal('c_action', action)
            self.record_signal('c_reward', reward)
            self.record_signal('c_value_reward', self.state['x_value_reward'])
            self.record_signal('x_learning_rate_reward', self.state['x_learning_rate_reward'])
            self.record_signal('x_value_reward_not_chosen', self.state['x_value_reward'])
            self.record_signal('x_value_choice_chosen', self.state['x_value_choice'])
            self.record_signal('x_value_choice_not_chosen', self.state['x_value_choice'])
            
            # updates for x_value_reward
            learning_rate_reward = self.call_module(
                key_module='x_learning_rate_reward',
                key_state='x_learning_rate_reward',
                action=action,
                inputs=(reward, self.state['x_value_reward']),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
            )
            
            next_value_reward_chosen = self.call_module(
                key_module='x_value_reward_chosen',
                key_state='x_value_reward',
                action=action,
                inputs=(reward, learning_rate_reward),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                )
            
            next_value_reward_not_chosen = self.call_module(
                key_module='x_value_reward_not_chosen',
                key_state='x_value_reward',
                action=1-action,
                inputs=None,
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                )
            
            # updates for x_value_choice
            next_value_choice_chosen = self.call_module(
                key_module='x_value_choice_chosen',
                key_state='x_value_choice',
                action=action,
                inputs=None,
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
                )
            
            next_value_choice_not_chosen = self.call_module(
                key_module='x_value_choice_not_chosen',
                key_state='x_value_choice',
                action=1-action,
                inputs=None,
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
                )
            
            # updating the memory state
            self.state['x_learning_rate_reward'] = learning_rate_reward
            self.state['x_value_reward'] = next_value_reward_chosen + next_value_reward_not_chosen
            self.state['x_value_choice'] = next_value_choice_chosen + next_value_choice_not_chosen
            
            # Now keep track of the logit in the output array
            logits[timestep] = self.state['x_value_reward'] * beta_reward + self.state['x_value_choice'] * beta_choice
            
        # post-process the forward pass; give here as inputs the logits, batch_first and all values from the memory state
        logits = self.post_forward_pass(logits, batch_first)
        
        return logits, self.get_state()
    

WEINHARDT_2025_CONFIG = SpiceConfig(
    rnn_modules=['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen'],
    control_parameters=['c_action', 'c_reward_chosen', 'c_value_reward', 'c_value_choice'],
    # The new module which handles the not-chosen value, does not need any additional inputs except for the value
    library_setup = {
        # 'x_value_reward_chosen': ['c_reward'] -> Remove this one from the library as we are not going to identify the dynamics of a hard-coded equation
        'x_learning_rate_reward': ['c_reward_chosen', 'c_value_reward', 'c_value_choice'],
        'x_value_reward_not_chosen': ['c_reward_chosen', 'c_value_choice'],
        'x_value_choice_chosen': ['c_value_reward'],
        'x_value_choice_not_chosen': ['c_value_reward'],
    },

    # Further, the new module should be applied only to the not-chosen values
    filter_setup = {
        'x_learning_rate_reward': ['c_action', 1, True],
        'x_value_reward_not_chosen': ['c_action', 0, True],
        'x_value_choice_chosen': ['c_action', 1, True],
        'x_value_choice_not_chosen': ['c_action', 0, True],
    }    
)


class Weinhardt2025RNN(BaseRNN):
    
    init_values = {
            'x_value_reward': 0.5,
            'x_value_choice': 0.,
            'x_learning_rate_reward': 0.,
        }
    
    def __init__(
        self,
        n_actions: int,
        n_participants: int,
        embedding_size = 32,
        dropout = 0.5,
        leaky_relu = 0.01,
        device = torch.device('cpu'),
        **kwargs,
    ):
        
        super().__init__(n_actions=n_actions, device=device, n_participants=n_participants, embedding_size=embedding_size)
                
        # set up the participant-embedding layer
        self.participant_embedding = self.setup_embedding(num_embeddings=n_participants, embedding_size=embedding_size)
        
        # scaling factor (inverse noise temperature) for each participant for the values which are handled by an hard-coded equation
        self.betas['x_value_reward'] = self.setup_constant(embedding_size=self.embedding_size, leaky_relu=leaky_relu)
        self.betas['x_value_choice'] = self.setup_constant(embedding_size=self.embedding_size, leaky_relu=leaky_relu)
        
        # set up the submodules
        self.submodules_rnn['x_learning_rate_reward'] = self.setup_module(input_size=3+self.embedding_size)
        self.submodules_rnn['x_value_reward_not_chosen'] = self.setup_module(input_size=2+self.embedding_size)
        self.submodules_rnn['x_value_choice_chosen'] = self.setup_module(input_size=1+self.embedding_size)
        self.submodules_rnn['x_value_choice_not_chosen'] = self.setup_module(input_size=1+self.embedding_size)
        
        # set up hard-coded equations
        self.submodules_eq['x_value_reward_chosen'] = self.x_value_reward_chosen
    
    def x_value_reward_chosen(self, value, inputs):
        return value + inputs[..., 1] * (inputs[..., 0] - value)
    
    def forward(self, inputs, prev_state=None, batch_first=False):
        """Forward pass of the RNN

        Args:
            inputs (torch.Tensor): includes all necessary inputs (action, reward, participant id) to the RNN to let it compute the next action
            prev_state (Tuple[torch.Tensor], optional): That's the previous memory state of the RNN containing the reward-based value. Defaults to None.
            batch_first (bool, optional): Indicates whether the first dimension of inputs is batch (True) or timesteps (False). Defaults to False.
        """
        
        # First, we have to initialize all the inputs and outputs (i.e. logits)
        input_variables, ids, logits, timesteps = self.init_forward_pass(inputs, prev_state, batch_first)
        actions, rewards, _, _ = input_variables
        participant_id, _ = ids
        
        # derive more observations
        rewards_chosen = (actions * rewards).sum(dim=-1, keepdim=True).repeat(1, 1, self._n_actions)
        # rewards_not_chosen = ((1-actions) * rewards).sum(dim=-1, keepdim=True).repeat(1, 1, self._n_actions)
        
        # Here we compute now the participant embeddings for each entry in the batch
        participant_embedding = self.participant_embedding(participant_id[:, 0].int())
        
        for timestep, action, reward_chosen in zip(timesteps, actions, rewards_chosen): #, rewards_not_chosen
            
            # record the current memory state and control inputs to the modules for SINDy training
            if not self.training and len(self.submodules_sindy)==0:
                self.record_signal('c_action', action)
                self.record_signal('c_reward_chosen', reward_chosen)
                # self.record_signal('c_reward_not_chosen', reward_not_chosen)
                self.record_signal('c_value_reward', self.state['x_value_reward'])
                self.record_signal('c_value_choice', self.state['x_value_choice'])
                self.record_signal('x_learning_rate_reward', self.state['x_learning_rate_reward'])
                self.record_signal('x_value_reward_not_chosen', self.state['x_value_reward'])
                self.record_signal('x_value_choice_chosen', self.state['x_value_choice'])
                self.record_signal('x_value_choice_not_chosen', self.state['x_value_choice'])
            
            # updates for x_value_reward
            learning_rate_reward = self.call_module(
                key_module='x_learning_rate_reward',
                key_state='x_learning_rate_reward',
                action=action,
                inputs=(
                    reward_chosen, 
                    # reward_not_chosen, 
                    self.state['x_value_reward'], 
                    self.state['x_value_choice'],
                    ),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
            )
            
            next_value_reward_chosen = self.call_module(
                key_module='x_value_reward_chosen',
                key_state='x_value_reward',
                action=action,
                inputs=(
                    reward_chosen, 
                    learning_rate_reward,
                    ),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                )

            next_value_reward_not_chosen = self.call_module(
                key_module='x_value_reward_not_chosen',
                key_state='x_value_reward',
                action=1-action,
                inputs=(
                    reward_chosen, 
                    # reward_not_chosen, 
                    self.state['x_value_choice'],
                    ),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
                )
            
            # updates for x_value_choice
            next_value_choice_chosen = self.call_module(
                key_module='x_value_choice_chosen',
                key_state='x_value_choice',
                action=action,
                inputs=(self.state['x_value_reward']),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
                )
            
            next_value_choice_not_chosen = self.call_module(
                key_module='x_value_choice_not_chosen',
                key_state='x_value_choice',
                action=1-action,
                inputs=(self.state['x_value_reward']),
                participant_embedding=participant_embedding,
                participant_index=participant_id,
                activation_rnn=torch.nn.functional.sigmoid,
                )
            
            # updating the memory state
            self.state['x_learning_rate_reward'] = learning_rate_reward
            self.state['x_value_reward'] = next_value_reward_chosen + next_value_reward_not_chosen
            self.state['x_value_choice'] = next_value_choice_chosen + next_value_choice_not_chosen
             
            # Now keep track of the logit in the output array
            logits[timestep] = self.state['x_value_reward'] * self.betas['x_value_reward'](participant_embedding) + self.state['x_value_choice'] * self.betas['x_value_choice'](participant_embedding)
            
        # post-process the forward pass; give here as inputs the logits, batch_first and all values from the memory state
        logits = self.post_forward_pass(logits, batch_first)
                
        return logits, self.get_state()


import numpy as np
from sklearn.metrics import log_loss, mean_squared_error
from typing import Iterable, List, Dict, Tuple, Callable, Union
import itertools
import random
from tqdm import tqdm
import dill
import torch

from .bandits import AgentNetwork, AgentSpice, get_update_dynamics, BanditSession, BanditsDrift
from .rnn_utils import DatasetRNN
from .model_evaluation import log_likelihood


def make_sindy_data(
    dataset,
    agent,
    sessions=-1,
    ):

  # Get training data for SINDy
  # put all relevant signals in x_train

  if not isinstance(sessions, Iterable) and sessions == -1:
    # use all sessions
    sessions = np.arange(len(dataset))
  else:
    # use only the specified sessions
    sessions = np.array(sessions)
    
  n_control = 2
  
  choices = np.stack([dataset[i].choices for i in sessions], axis=0)
  rewards = np.stack([dataset[i].rewards for i in sessions], axis=0)
  qs = np.stack([dataset[i].q for i in sessions], axis=0)
  
  choices_oh = np.zeros((len(sessions), choices.shape[1], agent._n_actions))
  for sess in sessions:
    # one-hot encode choices
    choices_oh[sess] = np.eye(agent._n_actions)[choices[sess]]
    
  # concatenate all qs values of one sessions along the trial dimension
  qs_all = np.concatenate([np.stack([np.expand_dims(qs_sess[:, i], axis=-1) for i in range(agent._n_actions)], axis=0) for qs_sess in qs], axis=0)
  c_all = np.concatenate([np.stack([c_sess[:, i] for i in range(agent._n_actions)], axis=0) for c_sess in choices_oh], axis=0)
  r_all = np.concatenate([np.stack([r_sess for _ in range(agent._n_actions)], axis=0) for r_sess in rewards], axis=0)
  
  # get observed dynamics
  x_train = qs_all
  feature_names = ['q']

  # get control
  control_names = []
  control = np.zeros((*x_train.shape[:-1], n_control))
  control[:, :, 0] = c_all
  control_names += ['c']
  control[:, :, n_control-1] = r_all
  control_names += ['r']
  
  feature_names += control_names
  
  print(f'Shape of Q-Values is: {x_train.shape}')
  print(f'Shape of control parameters is: {control.shape}')
  print(f'Feature names are: {feature_names}')
  
  # make x_train and control sequences instead of arrays
  x_train = [x_train_sess for x_train_sess in x_train]
  control = [control_sess for control_sess in control]
 
  return x_train, control, feature_names


def create_dataset(
  agent: AgentNetwork,
  data: DatasetRNN,
  rnn_modules: List,
  control_signals: List,
  dataprocessing: Dict[str, List] = None,
  shuffle: bool = False,
  groupby: str = 'c_action',
  ):
  
  n_trials = data.xs.shape[1]
  keys_x = rnn_modules
  
  keys_c = control_signals
  
  x_train = {key: [] for key in keys_x}
  control = {key: [] for key in keys_c}
  
  # determine whether trimming is specified in any of the variables of the dataprocessing-setup
  if dataprocessing is not None and any([int(dataprocessing[key][0]) for key in dataprocessing]):
    trimming = int(0.25*n_trials)
  else:
    trimming = 0
  
  for session in data.xs:
    # perform agent updates to record values over trials
    agent = get_update_dynamics(session, agent)[-1]
    
    # sort the data of one session into the corresponding signals
    for key in keys_x+keys_c:
      if len(agent._model.get_recording(key)) > 0:
        for index_action in range(agent._n_actions):
          # get all recorded values for the current session of one specific key 
          recording = agent._model.get_recording(key)
          # create tensor from list of tensors 
          values = np.concatenate(recording)[trimming:, index_action]
          # remove insignificant updates with a high-pass filter: check if dv/dt > threshold; otherwise set v(t=1) = v(t=0)
          # dvdt = np.abs(np.diff(values))
          # for index_time in range(1, dvdt.shape[0]):
          #   values[index_time] = np.where(dvdt[index_time-1] > highpass_threshold_dt, values[index_time], values[index_time-1])
          
          if key in keys_x:
            x_train[key].append(values)
          elif key in keys_c:
            control[key].append(values)
          
  feature_names = None
  
  # group recorded values into sequences by grouping them such that the groupby values do not change for one sequence
  # -> e.g. only c_action = 1.0 in one sequence; only c_action = 0.0 in next sequence (useful to split into chosen and not chosen)
  # Use itertools.groupby to find the indices of groups in the target sequence
  # groups = []
  grouped_x_train = []
  grouped_control = []
  for index_session, groupby_session in enumerate(control[groupby]):
    for key_group, group in itertools.groupby(enumerate(groupby_session), key=lambda x: x[1]):
      indices_group = [g[0] for g in list(group)]
      grouped_x_train_array = np.zeros((len(indices_group), len(keys_x)))
      grouped_control_array = np.zeros((len(indices_group), len(keys_c)))
      for index_key, key in enumerate(keys_x):
        grouped_x_train_array[:, index_key] = x_train[key][index_session][indices_group]
      for index_key, key in enumerate(keys_c):
        grouped_control_array[:, index_key] = control[key][index_session][indices_group]
      grouped_x_train.append(grouped_x_train_array)
      grouped_control.append(grouped_control_array)
    
  # data processing  
  if dataprocessing is not None:
    for index_key, key in enumerate(keys_x):
      # Offset-clearing
      if key in dataprocessing.keys() and int(dataprocessing[key][1]):
        x_min = np.min(np.concatenate(grouped_x_train)[:, index_key])
        for index_group in range(len(grouped_x_train)):
          grouped_x_train[index_group][:, index_key] -= x_min
      # Normalization
      if key in dataprocessing.keys() and int(dataprocessing[key][2]):
        x_min = np.min(np.concatenate(grouped_x_train)[:, index_key])
        x_max = np.max(np.concatenate(grouped_x_train)[:, index_key])
        for index_group in range(len(grouped_x_train)):
          grouped_x_train[index_group][:, index_key] = (grouped_x_train[index_group][:, index_key] - x_min) / (x_max - x_min)
          
    for index_key, key in enumerate(keys_c):
      # Offset-clearing
      if key in dataprocessing.keys() and int(dataprocessing[key][1]):
        x_min = np.min(np.concatenate(grouped_control)[:, index_key])
        for index_group in range(len(grouped_control)):
          grouped_control[index_group][:, index_key] -= x_min
      # Normalization
      if key in dataprocessing.keys() and int(dataprocessing[key][2]):
        x_min = np.min(np.concatenate(grouped_control)[:, index_key])
        x_max = np.max(np.concatenate(grouped_control)[:, index_key])
        for index_group in range(len(grouped_control)):
          grouped_control[index_group][:, index_key] = (grouped_control[index_group][:, index_key] - x_min) / (x_max - x_min)
  
    # compute scaling factor for beta
  #   beta_scaling = np.abs(x_maxs - x_mins).reshape(-1)
  # else:
  beta_scaling = None#np.ones((x_train_array.shape[-1]))
  
  if shuffle:
    # shuffle_idx = np.random.permutation(len(grouped_x_train))
    # grouped_x_train = grouped_x_train[shuffle_idx]
    # grouped_control = grouped_control[shuffle_idx]
    zipped = list(zip(grouped_x_train, grouped_control))
    random.shuffle(zipped)
    grouped_x_train, grouped_control = zip(*zipped)
    grouped_x_train, grouped_control = list(grouped_x_train), list(grouped_control)
  
  return grouped_x_train, grouped_control, feature_names, beta_scaling


def check_library_setup(library_setup: Dict[str, List[str]], feature_names: List[str], verbose=False) -> bool:
  msg = '\n'
  for key in library_setup.keys():
    if key not in feature_names:
      msg += f'Key {key} not in feature_names.\n'
    else:
      for feature in library_setup[key]:
        if feature not in feature_names:
          msg += f'Key {key}: Feature {feature} not in feature_names.\n'
  if msg != '\n':
    msg += f'Valid feature names are {feature_names}.\n'
    print(msg)
    return False
  else:
    if verbose:
      print('Library setup is valid. All keys and features appear in the provided list of features.')
    return True
  

def remove_control_features(control_variables: List[np.ndarray], feature_names: List[str], target_feature_names: List[str]) -> List[np.ndarray]:
  if len(target_feature_names) > 0:
    index_target_features = []
    for index_f, f_name in enumerate(feature_names):
      if f_name in target_feature_names:
        index_target_features.append(index_f)
    index_target_features = np.array(index_target_features)
    
    for index_group in range(len(control_variables)):
      control_variables[index_group] = control_variables[index_group][:, index_target_features]
  else:
    # for index_group in range(len(control_variables)):
    #   control_variables[index_group] = np.zeros_like(control_variables[index_group][:, :1])
    return None
  return control_variables

def conditional_filtering(x_train: List[np.ndarray], control: List[np.ndarray], feature_names: List[str], feature_filter: str, condition: float, remove_feature_filter=True) -> Tuple[List[np.ndarray], List[np.ndarray]]:
  x_filtered = []
  control_filtered = []
  index_filter = feature_names.index(feature_filter)-1
  for index_group in range(len(x_train)):
    if control[index_group][0, index_filter] == condition:
      x_filtered.append(x_train[index_group])
      control_filtered.append(control[index_group])

  if remove_feature_filter:
    feature_names.pop(index_filter+1)
    for index_group in range(len(x_filtered)):
      control_filtered[index_group] = control_filtered[index_group][:, np.arange(len(feature_names))!=index_filter]
    
    
  return x_filtered, control_filtered, feature_names


def sindy_loss_x(agent: Union[AgentSpice, AgentNetwork], data: List[BanditSession], loss_fn: Callable = log_loss):
  """Compute the loss of the SINDy model directly on the data in x-coordinates i.e. predicting behavior.
  This loss is not used for SINDy-Training, but for analysis purposes only.

  Args:
      model (ps.SINDy): _description_
      x_data (DatasetRNN): _description_
      loss_fn (Callable, optional): _description_. Defaults to log_loss.
  """
  
  loss_total = 0
  for experiment in data:
    agent.new_sess()
    choices = experiment.choices
    rewards = experiment.rewards
    loss_session = 0
    for t in range(len(choices)-1):
      beta = agent._beta if hasattr(agent, "_beta") else 1
      y_pred = np.exp(agent.q * beta)/np.sum(np.exp(agent.q * beta))
      agent.update(choices[t], rewards[t])
      loss_session += loss_fn(np.eye(agent._n_actions)[choices[t+1]], y_pred)
    loss_total += loss_session/(t+1)
  return loss_total/len(data)


def bandit_loss(agent: Union[AgentSpice, AgentNetwork], data: List[BanditSession], coordinates: str = "x"):
  """Compute the loss of the SINDy model directly on the data in z-coordinates i.e. predicting q-values.
  This loss is also used for SINDy-Training.

  Args:
      model (ps.SINDy): _description_
      x_data (DatasetRNN): _description_
      loss_fn (Callable): _description_. Defaults to log_loss.
      coordinates (str): Defines the coordinate system in which to compute the loss. Can be either "x" (predicting behavior) or "z" (comparing choice probabilities). Defaults to "x".
      """
  
  loss_total = 0
  for experiment in data:
    agent.new_sess()
    choices = experiment.choices
    rewards = experiment.rewards
    qs = np.exp(experiment.q)/np.sum(np.exp(experiment.q))
    loss_session = 0
    for t in range(len(choices)-1):
      beta = agent._beta if hasattr(agent, "_beta") else 1
      y_pred = np.exp(agent.q * beta)/np.sum(np.exp(agent.q * beta))
      agent.update(choices[t], rewards[t])
      if coordinates == 'x':
        y_target = np.eye(agent._n_actions)[choices[t+1]]
        loss_session = log_loss(y_target, y_pred)
      elif coordinates == 'z':
        y_target = qs[t]
        loss_session = mean_squared_error(y_target, y_pred)
    loss_total += loss_session/(t+1)
  return loss_total/len(data)


def remove_bad_participants(agent_spice: AgentSpice, agent_rnn: AgentNetwork, dataset: DatasetRNN, participant_ids: Iterable[int], trial_likelihood_difference_threshold: float=0.05, verbose: bool = False):
  """Check for badly fitted participants in the SPICE models w.r.t. the SPICE-RNN and return only the IDs of the well-fitted participants.

  Args:
      agent_spice (AgentSpice): _description_
      agent_rnn (AgentNetwork): _description_
      dataset_test (DatasetRNN): _description_
      participant_ids (Iterable[int]): _description_
      verbose (bool, optional): _description_. Defaults to False.

  Returns:
      AgentSpice: SPICE agent without badly fitted participants
      Iterable[int]: List of well-fitted participants
  """
  # if verbose:
  print("\nFiltering badly fitted SPICE models...")
  
  filtered_participant_ids = []
  
  # Create a copy of the valid participant IDs
  valid_participant_ids = list(participant_ids)
  
  removed_pids = []
  for pid in tqdm(participant_ids):
      # Skip if participant is not in the SPICE model
      if pid not in agent_spice._model.submodules_sindy[list(agent_spice._model.submodules_sindy.keys())[0]]:
          continue
         
      # Calculate normalized log likelihood for SPICE and RNN
      mask_participant_id = dataset.xs[:, 0, -1] == pid
      if not mask_participant_id.any():
          continue
          
      participant_data = DatasetRNN(*dataset[mask_participant_id])
      
      # Get probabilities from SPICE and RNN models
      agent_spice.new_sess(participant_id=pid)
      agent_rnn.new_sess(participant_id=pid)
      
      # Calculate NLL for both models
      probs_spice = get_update_dynamics(experiment=participant_data.xs, agent=agent_spice)[1]
      probs_rnn = get_update_dynamics(experiment=participant_data.xs, agent=agent_rnn)[1]
      n_trials_test = len(probs_spice)
      
      # Calculate scores (negative log likelihood)
      ll_spice = log_likelihood(data=participant_data.ys[0, :n_trials_test].cpu().numpy(), probs=probs_spice)
      ll_rnn = log_likelihood(data=participant_data.ys[0, :n_trials_test].cpu().numpy(), probs=probs_rnn)
      
      # Check if SPICE model is badly fitted
      spice_per_action_likelihood = np.exp(ll_spice/(n_trials_test*agent_rnn._n_actions))
      rnn_per_action_likelihood = np.exp(ll_rnn/(n_trials_test*agent_rnn._n_actions))
      
      # Idea for filter criteria:
      # If accuracy is very low for SPICE (near chance) but not so low for RNN then bad SPICE fitting (at least a bit higher than chance)
      # TODO: Check for better filter criteria
      if rnn_per_action_likelihood - spice_per_action_likelihood > trial_likelihood_difference_threshold:
          if verbose:
              print(f'SPICE trial likelihood ({spice_per_action_likelihood:.2f}) is unplausibly low compared to RNN trial likelihood ({rnn_per_action_likelihood:.2f}).')
              print(f'SPICE optimizer may be badly parameterized. Skipping participant {pid}.')
          
          # Remove this participant from the SPICE model
          for module in agent_spice._model.submodules_sindy:
              if pid in agent_spice._model.submodules_sindy[module]:
                  del agent_spice._model.submodules_sindy[module][pid]
          
          # Remove from valid participant IDs
          if pid in valid_participant_ids:
              valid_participant_ids.remove(pid)
              removed_pids.append(pid)
      else:
          # Keep track of filtered (good) participants
          filtered_participant_ids.append(pid)
  
  if verbose:
      print(f"\nAfter filtering: {len(valid_participant_ids)} of {len(participant_ids)} participants have well-fitted SPICE models.")
      print(f"Removed participants: {removed_pids}")
  
  return agent_spice, np.array(valid_participant_ids)


def save_spice(agent_spice: AgentSpice, file: str):
  """Saves the SINDy models (coefficients) in a pickle file
  
  Args:
      file (str): file where to store the SINDy coefficients
  """  
    
  with open(file, 'wb') as f:
    dill.dump(agent_spice.get_modules(), f)
    
    
def load_spice(file) -> Dict:
  with open(file, 'rb') as f:
    spice_modules = dill.load(f)
  return spice_modules


def generate_off_policy_data(participant_id: int, block: int, experiment_id: int, n_trials_off_policy: int, n_trials_same_action_off_policy: int, n_sessions_off_policy: int = 1, n_actions: int = 2, additional_inputs: np.ndarray = np.zeros(0), sigma_drift: float = 0.2) -> DatasetRNN:
  """Generates a simple off-policy dataset where each action is repeated n_trials_same_action_off_policy times and then switched.

  Args:
      participant_id (int): _description_
      n_trials_off_policy (int): _description_
      n_trials_same_action_off_policy (int): _description_
      n_sessions_off_policy (int, optional): _description_. Defaults to 1.
      n_actions (int, optional): _description_. Defaults to 2.

  Returns:
      DatasetRNN: Off-policy dataset
  """
  
  # set up environment to create an off-policy dataset (w.r.t to trained RNN) of arbitrary length
  # The trained RNN will then perform value updates to get off-policy data
  environment = BanditsDrift(sigma=sigma_drift, n_actions=n_actions)
  
  # create a dummy dataset where each choice is chosen for n times and then an action switch occures
  xs = torch.zeros((n_sessions_off_policy, n_trials_off_policy, 2*n_actions+3+additional_inputs.shape[-1])) - 1
  for session in range(n_sessions_off_policy):
      # initialize first action
      current_action = torch.zeros(n_actions)
      current_action[0] = 1
      for trial in range(n_trials_off_policy):
          current_action_index = torch.argmax(current_action).int().item()
          reward = torch.tensor(environment.step(current_action_index))
          xs[session, trial, :2*n_actions] = torch.concat((current_action, reward))
          # action switch - go to next possible action item and if final go to first one
          if trial >= n_trials_same_action_off_policy and trial % n_trials_same_action_off_policy == 0:
              current_action[current_action_index] = 0
              current_action[current_action_index+1 if current_action_index+1 < len(current_action) else 0] = 1
              
  # setup of dataset
  ys = xs[:, 1:, :n_actions]
  xs = xs[:, :-1]
  xs[..., 2*n_actions:-3] = additional_inputs
  xs[..., -3] = block
  xs[..., -2] = experiment_id
  xs[..., -1] = participant_id
  
  # dataset_fit = DatasetRNN(xs_fit, ys_fit)
  # # repeat the off-policy data for every participant and add the corresponding participant ID
  # xs_fit = dataset_fit.xs.repeat(len(participant_ids), 1, 1)
  # ys_fit = dataset_fit.ys.repeat(len(participant_ids), 1, 1)
  # set participant ids correctly
  # for index_pid in range(0, len(participant_ids)):
  #     xs[n_sessions_off_policy*index_pid:n_sessions_off_policy*(index_pid+1):, :, -1] = participant_ids[index_pid]
  
  return DatasetRNN(xs=xs, ys=ys)


SindyConfig = {
  
  # tracked variables and control signals in the RNN
  'rnn_modules': ['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen'],
  'control_parameters': ['c_action', 'c_reward', 'c_value_reward', 'c_value_choice'],
  
  # library setup: 
  # which terms are allowed as control inputs in each SINDy model
  # key is the SINDy model name, value is a list of allowed control inputs from the list of control signals 
  'library_setup': {
      'x_learning_rate_reward': ['c_reward', 'c_value_reward', 'c_value_choice'],
      'x_value_reward_not_chosen': ['c_value_choice'],
      'x_value_choice_chosen': ['c_value_reward'],
      'x_value_choice_not_chosen': ['c_value_reward'],
  },
  
  # data-filter setup: 
  # which samples are allowed as training samples in each SINDy model based on the given filter condition (conditions are always equality conditions)
  # key is the SINDy model name, value is a list with a triplet of values:
  #   1. str: feature name to be used as a filter
  #   2. numeric: the numeric filter condition
  #   3. bool: remove feature from control inputs --> TODO: check if this is necessary or makes things just more complicated
  # Multiple conditions can also be given as a list of triplets.
  # Example:
  #   'x_value_choice_not_chosen': ['c_action', 0, True] means that for the SINDy model 'x_value_choice_not_chosen', only samples where the feature 'c_action' == 0 are used for training the SINDy model. 
  #   The control parameter 'c_action' is removed afterwards from the list of control signals for training of the model
  'filter_setup': {
      'x_learning_rate_reward': ['c_action', 1, True],
      'x_value_reward_not_chosen': ['c_action', 0, True],
      'x_value_choice_chosen': ['c_action', 1, True],
      'x_value_choice_not_chosen': ['c_action', 0, True],
  },
  
  # data pre-processing setup:
  # define the processing steps for each variable and control signal.
  # possible processing steps are: 
  #   1. Trimming: Remove the first 25% of the samples along the time-axis. This is useful if the RNN begins with a variable at 0 but then accumulates first first to a specific default value, i.e. the range changes from (0, p) to (q, q+p). That way the data is cleared of the accumulation process. Trimming will be active for all variables, if it is active for one. 
  #   2. Offset-Clearing: Clearup any offset by determining the minimal value q of a variable and move the value range from (q, q+p) -> (0, p). This step makes SINDy equations less complex and aligns them more with RL-Theory
  #   3. Normalization: Scale the value range of a variable to x_max - x_min = 1. Offset-Clearing is recommended to achieve a value range of (0, 1) 
  # The processing steps are passed in the form of a binary triplet in this order: (Trimming, Offset-Clearing, Normalization) 
  'dataprocessing_setup': {
      'x_learning_rate_reward': [0, 0, 0],
      'x_value_reward_not_chosen': [0, 0, 0],
      'x_value_choice_chosen': [1, 1, 0],
      'x_value_choice_not_chosen': [1, 1, 0],
      'c_value_reward': [0, 0, 0],
      'c_value_choice': [1, 1, 0],
  },
  
}


SindyConfig_eckstein2022 = {
  
  # tracked variables and control signals in the RNN
  'rnn_modules': ['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen'],
  'control_parameters': ['c_action', 'c_reward_chosen', 'c_value_reward', 'c_value_choice'],
  
  # library setup: 
  # which terms are allowed as control inputs in each SINDy model
  # key is the SINDy model name, value is a list of allowed control inputs from the list of control signals 
  'library_setup': {
      'x_learning_rate_reward': ['c_reward_chosen', 'c_value_reward', 'c_value_choice'],
      'x_value_reward_not_chosen': ['c_reward_chosen', 'c_value_choice'],
      'x_value_choice_chosen': ['c_value_reward'],
      'x_value_choice_not_chosen': ['c_value_reward'],
  },
  
  # data-filter setup: 
  # which samples are allowed as training samples in each SINDy model based on the given filter condition (conditions are always equality conditions)
  # key is the SINDy model name, value is a list with a triplet of values:
  #   1. str: feature name to be used as a filter
  #   2. numeric: the numeric filter condition
  #   3. bool: remove feature from control inputs --> TODO: check if this is necessary or makes things just more complicated
  # Multiple conditions can also be given as a list of triplets.
  # Example:
  #   'x_value_choice_not_chosen': ['c_action', 0, True] means that for the SINDy model 'x_value_choice_not_chosen', only samples where the feature 'c_action' == 0 are used for training the SINDy model. 
  #   The control parameter 'c_action' is removed afterwards from the list of control signals for training of the model
  'filter_setup': {
      'x_learning_rate_reward': ['c_action', 1, True],
      'x_value_reward_not_chosen': ['c_action', 0, True],
      'x_value_choice_chosen': ['c_action', 1, True],
      'x_value_choice_not_chosen': ['c_action', 0, True],
  },
  
  # data pre-processing setup:
  # define the processing steps for each variable and control signal.
  # possible processing steps are: 
  #   1. Trimming: Remove the first 25% of the samples along the time-axis. This is useful if the RNN begins with a variable at 0 but then accumulates first first to a specific default value, i.e. the range changes from (0, p) to (q, q+p). That way the data is cleared of the accumulation process. Trimming will be active for all variables, if it is active for one. 
  #   2. Offset-Clearing: Clearup any offset by determining the minimal value q of a variable and move the value range from (q, q+p) -> (0, p). This step makes SINDy equations less complex and aligns them more with RL-Theory
  #   3. Normalization: Scale the value range of a variable to x_max - x_min = 1. Offset-Clearing is recommended to achieve a value range of (0, 1) 
  # The processing steps are passed in the form of a binary triplet in this order: (Trimming, Offset-Clearing, Normalization) 
  'dataprocessing_setup': {
      'x_learning_rate_reward': [0, 0, 0],
      'x_value_reward_not_chosen': [0, 0, 0],
      'x_value_choice_chosen': [0, 0, 0],
      'x_value_choice_not_chosen': [0, 0, 0],
      'c_value_reward': [0, 0, 0],
      'c_value_choice': [0, 0, 0],
  },
  
}


SindyConfig_dezfouli2019 = {
  
  # tracked variables and control signals in the RNN
  'rnn_modules': ['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen'],
  'control_parameters': ['c_action', 'c_reward_chosen', 'c_value_reward', 'c_value_choice'],
  
  # library setup: 
  # which terms are allowed as control inputs in each SINDy model
  # key is the SINDy model name, value is a list of allowed control inputs from the list of control signals 
  'library_setup': {
      'x_learning_rate_reward': ['c_reward_chosen', 'c_value_reward', 'c_value_choice'],
      'x_value_reward_not_chosen': ['c_reward_chosen', 'c_value_choice'],
      'x_value_choice_chosen': ['c_value_reward'],
      'x_value_choice_not_chosen': ['c_value_reward'],
  },
  
  # data-filter setup: 
  # which samples are allowed as training samples in each SINDy model based on the given filter condition (conditions are always equality conditions)
  # key is the SINDy model name, value is a list with a triplet of values:
  #   1. str: feature name to be used as a filter
  #   2. numeric: the numeric filter condition
  #   3. bool: remove feature from control inputs --> TODO: check if this is necessary or makes things just more complicated
  # Multiple conditions can also be given as a list of triplets.
  # Example:
  #   'x_value_choice_not_chosen': ['c_action', 0, True] means that for the SINDy model 'x_value_choice_not_chosen', only samples where the feature 'c_action' == 0 are used for training the SINDy model. 
  #   The control parameter 'c_action' is removed afterwards from the list of control signals for training of the model
  'filter_setup': {
      'x_learning_rate_reward': ['c_action', 1, True],
      'x_value_reward_not_chosen': ['c_action', 0, True],
      'x_value_choice_chosen': ['c_action', 1, True],
      'x_value_choice_not_chosen': ['c_action', 0, True],
  },
  
  # data pre-processing setup:
  # define the processing steps for each variable and control signal.
  # possible processing steps are: 
  #   1. Trimming: Remove the first 25% of the samples along the time-axis. This is useful if the RNN begins with a variable at 0 but then accumulates first first to a specific default value, i.e. the range changes from (0, p) to (q, q+p). That way the data is cleared of the accumulation process. Trimming will be active for all variables, if it is active for one. 
  #   2. Offset-Clearing: Clearup any offset by determining the minimal value q of a variable and move the value range from (q, q+p) -> (0, p). This step makes SINDy equations less complex and aligns them more with RL-Theory
  #   3. Normalization: Scale the value range of a variable to x_max - x_min = 1. Offset-Clearing is recommended to achieve a value range of (0, 1) 
  # The processing steps are passed in the form of a binary triplet in this order: (Trimming, Offset-Clearing, Normalization) 
  'dataprocessing_setup': {
      'x_learning_rate_reward': [0, 0, 0],
      'x_value_reward_not_chosen': [0, 0, 0],
      'x_value_choice_chosen': [0, 0, 0],
      'x_value_choice_not_chosen': [0, 0, 0],
      'c_value_reward': [0, 0, 0],
      'c_value_choice': [0, 0, 0],
  },
  
}


SindyConfig_dezfouli2019_blocks = {
  
  # tracked variables and control signals in the RNN
  'rnn_modules': ['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen', 'x_value_block', 'x_value_trial'],
  'control_parameters': ['c_action', 'c_reward_chosen', 'c_value_reward', 'c_value_choice', 'c_value_block', 'c_value_trial'],
  
  # library setup: 
  # which terms are allowed as control inputs in each SINDy model
  # key is the SINDy model name, value is a list of allowed control inputs from the list of control signals 
  'library_setup': {
      'x_learning_rate_reward': ['c_reward_chosen', 'c_value_reward', 'c_value_choice', 'c_value_block', 'c_value_trial'],
      'x_value_reward_not_chosen': ['c_reward_chosen', 'c_value_choice', 'c_value_block', 'c_value_trial'],
      'x_value_choice_chosen': ['c_value_reward', 'c_value_block', 'c_value_trial'],
      'x_value_choice_not_chosen': ['c_value_reward', 'c_value_block', 'c_value_trial'],
  },
  
  # data-filter setup: 
  # which samples are allowed as training samples in each SINDy model based on the given filter condition (conditions are always equality conditions)
  # key is the SINDy model name, value is a list with a triplet of values:
  #   1. str: feature name to be used as a filter
  #   2. numeric: the numeric filter condition
  #   3. bool: remove feature from control inputs --> TODO: check if this is necessary or makes things just more complicated
  # Multiple conditions can also be given as a list of triplets.
  # Example:
  #   'x_value_choice_not_chosen': ['c_action', 0, True] means that for the SINDy model 'x_value_choice_not_chosen', only samples where the feature 'c_action' == 0 are used for training the SINDy model. 
  #   The control parameter 'c_action' is removed afterwards from the list of control signals for training of the model
  'filter_setup': {
      'x_learning_rate_reward': ['c_action', 1, True],
      'x_value_reward_not_chosen': ['c_action', 0, True],
      'x_value_choice_chosen': ['c_action', 1, True],
      'x_value_choice_not_chosen': ['c_action', 0, True],
  },
  
  # data pre-processing setup:
  # define the processing steps for each variable and control signal.
  # possible processing steps are: 
  #   1. Trimming: Remove the first 25% of the samples along the time-axis. This is useful if the RNN begins with a variable at 0 but then accumulates first first to a specific default value, i.e. the range changes from (0, p) to (q, q+p). That way the data is cleared of the accumulation process. Trimming will be active for all variables, if it is active for one. 
  #   2. Offset-Clearing: Clearup any offset by determining the minimal value q of a variable and move the value range from (q, q+p) -> (0, p). This step makes SINDy equations less complex and aligns them more with RL-Theory
  #   3. Normalization: Scale the value range of a variable to x_max - x_min = 1. Offset-Clearing is recommended to achieve a value range of (0, 1) 
  # The processing steps are passed in the form of a binary triplet in this order: (Trimming, Offset-Clearing, Normalization) 
  'dataprocessing_setup': {
      'x_learning_rate_reward': [0, 0, 0],
      'x_value_reward_not_chosen': [0, 0, 0],
      'x_value_choice_chosen': [0, 0, 0],
      'x_value_choice_not_chosen': [0, 0, 0],
      'c_value_reward': [0, 0, 0],
      'c_value_choice': [0, 0, 0],
  },
  
}


#@title Import libraries
import sys
import os

import time
import torch
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable
import argparse
from typing import List

# warnings.filterwarnings("ignore")

# RL libraries
sys.path.append('resources')  # add source directoy to path
from resources import rnn, rnn_training, bandits, rnn_utils
from utils import convert_dataset, plotting

def main(
  checkpoint = False,
  model: str = None,
  data: str = None,
  class_rnn: type = None,

  # rnn parameters
  hidden_size = 8,
  embedding_size = 32,
  dropout = 0.5,

  # data and training parameters
  epochs = 128,
  train_test_ratio = 1.,
  l2_weight_decay=0,
  bagging = False,
  sequence_length = -1, # -1 for full sequence
  n_steps = -1,  # -1 for full sequence
  batch_size = -1,  # -1 for one batch per epoch
  learning_rate = 1e-2,
  convergence_threshold = 0,
  scheduler = False,
  additional_inputs_data: List[str] = None,

  # ground truth parameters
  n_trials = 200,
  n_sessions = 256,
  beta_reward = 3.,
  alpha_reward = 0.25,
  alpha_penalty = -1.,
  alpha_counterfactual = 0.,
  beta_choice = 0.,
  alpha_choice = 0.,
  forget_rate = 0.,
  confirmation_bias = 0.,
  parameter_variance = 0.,
  reward_prediction_error: Callable = None,
  
  # environment parameters
  n_actions = 2,
  sigma = 0.1,
  counterfactual = False,
  
  analysis: bool = False,
  participant_id: int = 0,
  save_checkpoints: bool = False,
  ):
  
  # print cuda devices available
  print(f'Cuda available: {torch.cuda.is_available()}')
  device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
  # device = torch.device('cpu')
  
  if not os.path.exists('params'):
    os.makedirs('params')

  if participant_id is None:
    participant_id = 0
  
  
  dataset_test = None
  agent = None
  if data is None:
    print('No path to dataset provided.')
    
    # setup
    environment = bandits.BanditsDrift(sigma=sigma, n_actions=n_actions, counterfactual=counterfactual)
    # environment = bandits.EnvironmentBanditsSwitch(sigma, counterfactual=counterfactual)
    agent = bandits.AgentQ(
      n_actions=n_actions, 
      beta_reward=beta_reward, 
      alpha_reward=alpha_reward, 
      alpha_penalty=alpha_penalty, 
      forget_rate=forget_rate, 
      confirmation_bias=confirmation_bias, 
      beta_choice=beta_choice, 
      alpha_choice=alpha_choice,
      alpha_counterfactual_reward=alpha_counterfactual,
      alpha_counterfactual_penalty=alpha_counterfactual, 
      parameter_variance=parameter_variance,
      )
    if reward_prediction_error is not None:
      agent.set_reward_prediction_error(reward_prediction_error)
    print('Setup of the environment and agent complete.')
    
    print('Generating the synthetic dataset...')
    dataset, _, _ = bandits.create_dataset(
        agent=agent,
        environment=environment,
        n_trials=n_trials,
        n_sessions=n_sessions,
        sample_parameters=parameter_variance!=0,
        sequence_length=sequence_length,
        device=device,
        )
    
    # set participant ids to 0
    dataset.xs[..., -1] = 0.
    
    if train_test_ratio == 1:
      dataset_test, _, _ = bandits.create_dataset(
        agent=agent,
        environment=environment,
        n_trials=256,
        n_sessions=n_sessions,
        sample_parameters=parameter_variance!=0,
        device=device)
      
      dataset_test.xs[..., -1] = 0.

    print('Generation of dataset complete.')
  else:
    dataset, _, df, _ = convert_dataset.convert_dataset(data, sequence_length=sequence_length, device=device, additional_inputs=additional_inputs_data)
    # dataset_test = rnn_utils.DatasetRNN(dataset.xs, dataset.ys)
    
    # check if groundtruth parameters in data - only applicable to generated data with e.g. utils/create_dataset.py
    if 'mean_beta_reward' in df.columns:
      # get parameters from dataset
      agent = bandits.AgentQ(
        beta_reward = df['beta_reward'].values[(df['session']==participant_id).values][0],
        alpha_reward = df['alpha_reward'].values[(df['session']==participant_id).values][0],
        alpha_penalty = df['alpha_penalty'].values[(df['session']==participant_id).values][0],
        confirmation_bias = df['confirmation_bias'].values[(df['session']==participant_id).values][0],
        forget_rate = df['forget_rate'].values[(df['session']==participant_id).values][0],
        beta_choice = df['beta_choice'].values[(df['session']==participant_id).values][0],
        alpha_choice = df['alpha_choice'].values[(df['session']==participant_id).values][0],
      )
      
  n_participants = len(dataset.xs[..., -1].unique())
  
  if isinstance(train_test_ratio, float):
    if train_test_ratio < 1:
      dataset_train, dataset_test = rnn_utils.split_data_along_timedim(dataset, train_test_ratio)
      dataset_train = bandits.DatasetRNN(dataset_train.xs, dataset_train.ys, sequence_length=sequence_length, device=device)
    else:
      dataset_train = bandits.DatasetRNN(dataset.xs, dataset.ys, sequence_length=sequence_length, device=device)
      # if dataset_test is None:
      #   dataset_test = bandits.DatasetRNN(dataset.xs, dataset.ys, device=device)
  elif isinstance(train_test_ratio, list):
    dataset_train, dataset_test = rnn_utils.split_data_along_sessiondim(dataset=dataset, list_test_sessions=train_test_ratio, device=device)
  elif train_test_ratio is None:
    dataset_train, dataset_test = dataset, dataset
  else:
    raise TypeError("train_test_raio must be either a float number or a list of integers containing the session/block ids which should be used as test sessions/blocks")
  
  if data is None and model is None:
    params_path = rnn_utils.parameter_file_naming(
      'params/params', 
      alpha_reward=alpha_reward, 
      beta_reward=beta_reward, 
      alpha_counterfactual=alpha_counterfactual,
      forget_rate=forget_rate, 
      beta_choice=beta_choice,
      alpha_choice=alpha_choice,
      alpha_penalty=alpha_penalty,
      confirmation_bias=confirmation_bias, 
      variance=parameter_variance, 
      verbose=True,
      )
  elif data is not None and model is None:
    params_path = 'params/params_' + data.split('/')[-1].replace('.csv', '.pkl')
  else:
    params_path = '' + model

  # define model
  if class_rnn is None:
    class_rnn = rnn.RLRNN
  model = class_rnn(
      n_actions=n_actions, 
      hidden_size=hidden_size, 
      embedding_size=embedding_size,
      dropout=dropout,
      n_participants=n_participants,
      ).to(device)
  
  optimizer_rnn = torch.optim.Adam(model.parameters(), lr=learning_rate, weight_decay=l2_weight_decay)
  
  print('Setup of the RNN model complete.')

  if checkpoint:
      model, optimizer_rnn = rnn_utils.load_checkpoint(params_path, model, optimizer_rnn)
      print('Loaded model parameters.')

  loss_test = None
  if epochs > 0:
    start_time = time.time()
    
    #Fit the RNN
    print('Training the RNN...')
    model, optimizer_rnn, _ = rnn_training.fit_model(
        model=model,
        dataset_train=dataset_train,
        # dataset_test=None,
        # dataset_test=dataset_test,
        # dataset_test=dataset_train,
        optimizer=optimizer_rnn,
        convergence_threshold=convergence_threshold,
        epochs=epochs,
        batch_size=batch_size,
        bagging=bagging,
        n_steps=n_steps,
        scheduler=scheduler,
        path_save_checkpoints=params_path if save_checkpoints else None,
    )
        
    # save trained parameters
    state_dict = {'model': model.state_dict(), 'optimizer': optimizer_rnn.state_dict()}
    
    print('Training finished.')
    torch.save(state_dict, params_path)
    print(f'Saved RNN parameters to file {params_path}.')
    print(f'Training took {time.time() - start_time:.2f} seconds.')
  
  # validate model
  str_data = 'test' if dataset_test is not None else 'train'
  print(f'\nTesting the trained RNN on the', str_data, 'dataset...')
  model.eval()
  with torch.no_grad():
    _, _, loss_test = rnn_training.fit_model(
        model=model,
        dataset_train=dataset_test if dataset_test is not None else dataset_train,
    )
  
  # -----------------------------------------------------------
  # Analysis
  # -----------------------------------------------------------
  
  if analysis:
    # print(f'Betas of model: {(model._beta_reward.item(), model._beta_choice.item())}')
    # Synthesize a dataset using the fitted network
    agent_rnn = bandits.AgentNetwork(model_rnn=model, n_actions=n_actions)
    
    # get analysis plot
    if agent is not None:
      agents = {'groundtruth': agent, 'rnn': agent_rnn}
    else:
      agents = {'rnn': agent_rnn}

    # fig, axs = plotting.plot_session(agents, dataset_test.xs[participant_id] if dataset_test is not None else dataset_train.xs[participant_id])
    fig, axs = plotting.plot_session(agents, dataset.xs[participant_id])
    
    title_ground_truth = ''
    if agent is not None:
      title_ground_truth += r'GT: $\beta_{reward}=$'+str(np.round(agent._beta_reward, 2)) + r'; $\beta_{choice}=$'+str(np.round(agent._beta_choice, 2))
    betas = agent_rnn.get_betas()
    title_rnn = r'RNN: $\beta_{reward}=$'+str(np.round(betas['x_value_reward'], 2)) + r'; $\beta_{choice}=$'+str(np.round(betas['x_value_choice'], 2))
    fig.suptitle(title_ground_truth + '\n' + title_rnn)
    plt.show()
    
  return model, loss_test


if __name__=='__main__':
  
  parser = argparse.ArgumentParser(description='Trains a SPICE-RNN on behavioral data to uncover the underlying Q-Values via different cognitive modules.')
  
  
  parser.add_argument('--checkpoint', action='store_true', help='Whether to load a checkpoint')
  parser.add_argument('--model', type=str, default=None, help='Model name to load from and/or save to parameters of RNN')
  parser.add_argument('--data', type=str, default=None, help='Path to dataset')
  
  # RNN parameters
  parser.add_argument('--hidden_size', type=int, default=8, help='Hidden size of the RNN')
  parser.add_argument('--embedding_size', type=int, default=32, help='Participant embedding size of the RNN')
  parser.add_argument('--dropout', type=float, default=0.5, help='Dropout rate')

  # data and training parameters
  parser.add_argument('--scheduler', action='store_true', help='Whether to use a learning rate scheduler during training')
  parser.add_argument('--bagging', action='store_true', help='Whether to use bagging')
  parser.add_argument('--epochs', type=int, default=8192, help='Number of epochs for training')
  parser.add_argument('--n_steps', type=int, default=-1, help='Number of recurrent steps per training call; -1: Use whole sequence at once;')
  parser.add_argument('--batch_size', type=int, default=-1, help='Batch size; -1: Use whole dataset at once;')
  parser.add_argument('--sequence_length', type=int, default=-1, help='Length of training sequences; -1: Use whole sequence at once;')
  parser.add_argument('--learning_rate', type=float, default=1e-2, help='Learning rate of the RNN')
  parser.add_argument('--l2_weight_decay', type=float, default=0, help='Learning rate of the RNN')
  parser.add_argument('--convergence_threshold', type=float, default=0, help='Convergence threshold to early-stop training')
  parser.add_argument('--train_test_ratio', type=str, default="1.0", help='Ratio of training data; Can also be a comma-separated list of integeres to indicate testing sessions.')
  
  # Ground truth parameters
  parser.add_argument('--n_trials', type=int, default=200, help='Number of trials per session')
  parser.add_argument('--n_sessions', type=int, default=256, help='Number of sessions')
  parser.add_argument('--beta_reward', type=float, default=3, help='Beta parameter for the Q-learning update rule')
  parser.add_argument('--alpha_reward', type=float, default=0.25, help='Alpha parameter for the Q-learning update rule')
  parser.add_argument('--alpha_penalty', type=float, default=-1., help='Learning rate for negative outcomes; if -1: same as alpha')
  parser.add_argument('--forget_rate', type=float, default=0., help='Forget rate')
  parser.add_argument('--beta_choice', type=float, default=0., help='Beta parameter for the Q-learning update rule')
  parser.add_argument('--alpha_choice', type=float, default=1., help='Alpha parameter for the Q-learning update rule')
  parser.add_argument('--alpha_counterfactual', type=float, default=0., help='Alpha parameter for the Q-learning update rule')

  # Environment parameters
  parser.add_argument('--n_actions', type=int, default=2, help='Number of possible actions')
  parser.add_argument('--sigma', type=float, default=0.2, help='Drift rate of the reward probabilities')
  parser.add_argument('--counterfactual', action='store_true', help='Counterfactual experiment with full feedback (for chosen and not chosen options)')

  # Analysis parameters
  parser.add_argument('--analysis', action='store_true', help='Whether to perform visual analysis on one participant (keyword argument: participant_id)')
  parser.add_argument('--participant_id', type=int, default=None, help='Participant ID for visual analysis (keyword argument: analysis)')
  parser.add_argument('--save_checkpoints', action='store_true', help='Save a checkpoint after 2^n training steps with $n\in\mathbb\{N\}$.')
  
  args = parser.parse_args()  
  
  # currently fixed 
  class_rnn = rnn.RLRNN_eckstein2022
  
  # convert train_test_ratio to number of list of numbers
  if ',' in args.train_test_ratio:
    train_test_ratio = [int(x) for x in args.train_test_ratio.split(',')]
  else:
    train_test_ratio = float(args.train_test_ratio)
    
  main(
    checkpoint = args.checkpoint,
    model = args.model,
    data = args.data,
    class_rnn=class_rnn,
    
    # rnn parameters
    hidden_size = args.hidden_size,
    embedding_size = args.embedding_size,
    dropout = args.dropout,

    # data and training parameters
    epochs = args.epochs,
    train_test_ratio = train_test_ratio,
    n_trials = args.n_trials,
    n_sessions = args.n_sessions,
    bagging = args.bagging,
    sequence_length = args.sequence_length,
    n_steps = args.n_steps,
    batch_size = args.batch_size,
    learning_rate = args.learning_rate,
    convergence_threshold = args.convergence_threshold,
    scheduler = args.scheduler,
    l2_weight_decay=args.l2_weight_decay,
    
    # ground truth parameters
    beta_reward = args.beta_reward,
    alpha_reward = args.alpha_reward,
    alpha_penalty = args.alpha_penalty,
    alpha_counterfactual = args.alpha_counterfactual,
    beta_choice = args.beta_choice,
    alpha_choice = args.alpha_choice,
    forget_rate = args.forget_rate,
    
    # environment parameters
    n_actions = args.n_actions,
    sigma = args.sigma,
    counterfactual = args.counterfactual,
    
    # analysis parameters
    analysis = args.analysis,
    save_checkpoints=args.save_checkpoints,
    participant_id = args.participant_id,
    )
  


from typing import List, Dict, Tuple, Iterable
import numpy as np
from copy import deepcopy
from tqdm import tqdm
import pandas as pd

from .sindy_utils import remove_bad_participants
from .fit_sindy import fit_sindy_pipeline
from .rnn_utils import DatasetRNN,split_data_along_timedim, split_data_along_sessiondim
from .bandits import AgentNetwork, AgentSpice, get_update_dynamics
from .model_evaluation import bayesian_information_criterion as loss_metric, log_likelihood
from .optimizer_selection import optimize_for_participant


def module_pruning(agent_spice: AgentSpice, dataset: DatasetRNN, participant_ids: Iterable[int], verbose: bool = False):
    """Check for badly fitted participants in the SPICE models w.r.t. the SPICE-RNN and return only the IDs of the well-fitted participants.

    Args:
        agent_spice (AgentSpice): _description_
        dataset_test (DatasetRNN): _description_
        participant_ids (Iterable[int]): _description_
        verbose (bool, optional): _description_. Defaults to False.

    Returns:
        AgentSpice: pruned SPICE agent
    """
    if verbose:
        print("\nPruning unnecessary SPICE modules...")
    
    n_removed_modules = 0
    for pid in tqdm(participant_ids):
        # Skip if participant is not in the SPICE model
        if pid not in agent_spice._model.submodules_sindy[list(agent_spice._model.submodules_sindy.keys())[0]]:
            if verbose:
                print("Participant ID", pid, "not found in SPICE model. Skipped that participant.")
            continue
        
        mask_participant_id = dataset.xs[:, 0, -1] == pid
        additional_embedding_inputs = dataset.xs[mask_participant_id, 0, agent_spice._n_actions*2:-3]
        agent_spice.new_sess(participant_id=pid, additional_embedding_inputs=additional_embedding_inputs)
        keys_betas = agent_spice.get_betas().keys()
        if not any([np.abs(agent_spice.get_betas()[key]) < 1e-2 for key in keys_betas]):
            continue
        
        # Calculate normalized log likelihood for SPICE and RNN
        mask_participant_id = dataset.xs[:, 0, -1] == pid
        if not mask_participant_id.any():
            if verbose:
                print("Participant ID", pid, "not found in dataset. Skipped that participant.")
            continue
            
        participant_data = DatasetRNN(*dataset[mask_participant_id])
        
        # Get probabilities from SPICE and RNN models
        agent_spice.new_sess(participant_id=pid)
        
        # Calculate action probabilities for original model
        probs_spice = get_update_dynamics(experiment=participant_data.xs, agent=agent_spice)[1]
        # n_trials_test = len(probs_spice)
        
        # Calculate action probabilities for pruned models
        for module in agent_spice.get_modules():
            agent_spice_modified = deepcopy(agent_spice)
            module_modified = agent_spice_modified.get_modules()[module][pid]
            
            # set all coefficients of the currently investigated module to 0
            module_modified.optimizer.coef_ = np.zeros_like(module_modified.optimizer.coef_)

            # put modified module back into agent
            # agent_spice_modified._model.submodules_sindy[module][pid] = module_modified
            
            # get new action probabilities and compare to original ones
            probs_spice_modified = get_update_dynamics(experiment=participant_data.xs, agent=agent_spice_modified)[1]
            if np.sum(np.abs(probs_spice - probs_spice_modified)) == 0 and np.sum(agent_spice.get_modules()[module][pid].coefficients()) != 0:
                if verbose:
                    print("Found unnecessary module", module, "for participant", pid)
                agent_spice._model.submodules_sindy[module][pid].optimizer.coef_ = np.zeros_like(module_modified.optimizer.coef_)
                n_removed_modules += 1
                
    if verbose:
        print("Removed modules:", n_removed_modules)
        
    return agent_spice
        

def fit_spice(
    rnn_modules: List[np.ndarray],
    control_signals: List[np.ndarray], 
    agent_rnn: AgentNetwork,
    data: DatasetRNN = None,
    polynomial_degree: int = 2, 
    library_setup: Dict[str, List[str]] = {},
    filter_setup: Dict[str, Tuple[str, float]] = {},
    optimizer_type: str = "SR3_weighted_l1",
    optimizer_threshold: float = 0.05,
    optimizer_alpha: float = 0.1,
    participant_id: int = None,
    shuffle: bool = False,
    dataprocessing: Dict[str, List] = None,
    n_trials_off_policy: int = 1000,
    n_sessions_off_policy: int = 1,
    n_trials_same_action_off_policy: int = 5,
    train_test_ratio: float = 1.0,
    deterministic: bool = True,
    get_loss: bool = False,
    verbose: bool = False,
    use_optuna: bool = False,
    filter_bad_participants: bool = False,
    pruning: bool = False,
    optuna_threshold: float = 0.03,
    optuna_n_trials: int = 50,
    ) -> Tuple[AgentSpice, float]:
    """Fit a SPICE agent by replacing RNN modules with SINDy equations.

    Args:
        rnn_modules (List[np.ndarray]): List of RNN module names to be replaced with SINDy
        control_signals (List[np.ndarray]): List of control signal names
        agent (AgentNetwork): The trained RNN agent
        data (DatasetRNN, optional): Dataset for training/evaluation. Defaults to None.
        polynomial_degree (int, optional): Polynomial degree for SINDy. Defaults to 2.
        library_setup (Dict[str, List[str]], optional): Dictionary mapping features to library components. Defaults to {}.
        filter_setup (Dict[str, Tuple[str, float]], optional): Dictionary mapping features to filter conditions. Defaults to {}.
        optimizer_type (str, optional): Type of optimizer to use. Defaults to "SR3_L1".
        optimizer_threshold (float, optional): Threshold for optimizer. Defaults to 0.05.
        optimizer_alpha (float, optional): Alpha parameter for optimizer. Defaults to 1.
        participant_id (int, optional): Specific participant ID to process. Defaults to None.
        shuffle (bool, optional): Whether to shuffle the data. Defaults to False.
        dataprocessing (Dict[str, List], optional): Data processing options. Defaults to None.
        n_trials_off_policy (int, optional): Number of off-policy trials. Defaults to 2048.
        n_sessions_off_policy (int, optional): Number of off-policy sessions. Defaults to 1.
        deterministic (bool, optional): Whether to use deterministic mode. Defaults to True.
        get_loss (bool, optional): Whether to compute loss. Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
        use_optuna (bool, optional): Whether to use Optuna for optimizer selection. Defaults to False.

    Returns:
        Tuple[AgentSpice, float]: The SPICE agent and its loss
    """
    
    if isinstance(train_test_ratio, float):
        data = split_data_along_timedim(data, split_ratio=train_test_ratio)[0]
    elif isinstance(train_test_ratio, list) or isinstance(train_test_ratio, tuple):
        data = split_data_along_sessiondim(data, list_test_sessions=train_test_ratio)[0]
    
    if participant_id is not None:
        participant_ids = [participant_id]
    elif data is not None and participant_id is None:
        participant_ids = data.xs[..., -1].unique().int().cpu().numpy()
    else:
        raise ValueError("Either data or participant_id are required.")
        
    sindy_modules = {rnn_module: {} for rnn_module in rnn_modules}
    optuna_pids = []
    likelihoods_rnn = []
    likelihoods_spice_before_optuna = []
    likelihoods_spice_after_optuna = []
    for pid in tqdm(participant_ids):
        
        if data is not None:
            mask_participant_id = data.xs[:, 0, -1] == pid
            data_pid = DatasetRNN(*data[mask_participant_id])
        
        # just fit the SINDy modules with the given parameters 
        sindy_modules_id = fit_sindy_pipeline(
            participant_id=pid,
            agent=agent_rnn,
            data=data_pid,
            rnn_modules=rnn_modules,
            control_signals=control_signals,
            sindy_library_setup=library_setup,
            sindy_filter_setup=filter_setup,
            sindy_dataprocessing=dataprocessing,
            optimizer_type=optimizer_type,
            optimizer_alpha=optimizer_alpha,
            optimizer_threshold=optimizer_threshold,
            polynomial_degree=polynomial_degree,
            shuffle=shuffle,
            n_sessions_off_policy=n_sessions_off_policy,
            n_trials_off_policy=n_trials_off_policy,
            n_trials_same_action_off_policy=n_trials_same_action_off_policy,
            catch_convergence_warning=False,
            verbose=verbose,
        )
        
        spice_modules_id = {rnn_module: {} for rnn_module in rnn_modules}
        for rnn_module in rnn_modules:
            spice_modules_id[rnn_module][pid] = sindy_modules_id[rnn_module]
        agent_spice_id = AgentSpice(model_rnn=agent_rnn._model, sindy_modules=spice_modules_id, n_actions=agent_rnn._n_actions)
        
        if use_optuna:
            probs_rnn = get_update_dynamics(agent=agent_rnn, experiment=data_pid.xs[0])[1]
            probs_spice = get_update_dynamics(agent=agent_spice_id, experiment=data_pid.xs[0])[1]
            lik_rnn = np.exp(log_likelihood(data=data_pid.xs[0, :probs_rnn.shape[0], :agent_rnn._n_actions].numpy(), probs=probs_rnn) / len(probs_rnn))
            lik_spice_before_optuna = np.exp(log_likelihood(data=data_pid.xs[0, :probs_rnn.shape[0], :agent_rnn._n_actions].numpy(), probs=probs_spice) / len(probs_spice))
            
            # if np.abs(lik_rnn - lik_spice_before_optuna) > optuna_threshold or np.isnan(lik_spice_before_optuna):
            if np.abs(probs_rnn - probs_spice).mean() > optuna_threshold or np.isnan(lik_spice_before_optuna):
                print(f"Likelihoods before optuna fitting: RNN = {np.round(lik_rnn, 5)}; SPICE =  {np.round(lik_spice_before_optuna, 5)}; Diff = {np.round(lik_rnn-lik_spice_before_optuna, 5)}")
                likelihoods_rnn.append(np.round(lik_rnn, 5))
                likelihoods_spice_before_optuna.append(np.round(lik_spice_before_optuna, 5))
                
                # find the best optimizer configuration for this participant using optuna
                optuna_pids.append(pid)
                print(f"Using optuna to find a better set of pysindy parameters for participant {pid}...")
                # Find optimal optimizer and parameters for this participant
                sindy_config = optimize_for_participant(
                    participant_id=pid,
                    agent_rnn=agent_rnn,
                    data=data_pid,
                    metric_rnn=probs_rnn,
                    rnn_modules=rnn_modules,
                    control_signals=control_signals,
                    library_setup=library_setup,
                    filter_setup=filter_setup,
                    polynomial_degree=polynomial_degree,
                    optimizer_type=optimizer_type,
                    n_sessions_off_policy=n_sessions_off_policy,
                    n_trials_off_policy=n_trials_off_policy,
                    n_trials_same_action_off_policy=n_trials_same_action_off_policy,
                    threshold=optuna_threshold,
                    n_trials_optuna=optuna_n_trials,  # Adjust as needed
                    verbose=verbose
                )
                
                # optuna_optimizer_type = sindy_config["optimizer_type"]
                optuna_optimizer_alpha = sindy_config["optimizer_alpha"]
                optuna_optimizer_threshold = sindy_config["optimizer_threshold"]
                # optuna_n_sessions_off_policy = sindy_config["n_sessions_off_policy"]
                # optuna_n_trials_off_policy = sindy_config["n_trials_off_policy"]
                # optuna_n_trials_same_action_off_policy = sindy_config["n_trials_same_action_off_policy"]
                
                if verbose:
                    print(f"\nUsing optimized parameters for participant {pid}:")
                    # print(f"\tOptimizer type: {optuna_optimizer_type}")
                    print(f"\tAlpha: {optuna_optimizer_alpha}")
                    print(f"\tThreshold: {optuna_optimizer_threshold}")
                    # print(f"\tOff-polcy sessions: {optuna_n_sessions_off_policy}")
                    # print(f"\tOff-polcy trials: {optuna_n_trials_off_policy}")
                    # print(f"\tSame action in off-polcy trials: {optuna_n_trials_same_action_off_policy}")
                    
                sindy_modules_optuna = fit_sindy_pipeline(
                participant_id=pid,
                agent=agent_rnn,
                data=data_pid,
                rnn_modules=rnn_modules,
                control_signals=control_signals,
                sindy_library_setup=library_setup,
                sindy_filter_setup=filter_setup,
                sindy_dataprocessing=dataprocessing,
                # optimizer_type=optuna_optimizer_type,
                optimizer_type=optimizer_type,
                optimizer_alpha=optuna_optimizer_alpha,
                optimizer_threshold=optuna_optimizer_threshold,
                polynomial_degree=polynomial_degree,
                shuffle=shuffle,
                # n_sessions_off_policy=optuna_n_sessions_off_policy,
                n_sessions_off_policy=n_sessions_off_policy,
                n_trials_off_policy=n_trials_off_policy,
                n_trials_same_action_off_policy=n_trials_same_action_off_policy,
                catch_convergence_warning=False,
                verbose=verbose,
                )
                
                spice_modules_optuna = {rnn_module: {} for rnn_module in rnn_modules}
                for rnn_module in rnn_modules:
                    spice_modules_optuna[rnn_module][pid] = sindy_modules_optuna[rnn_module]
                agent_spice_optuna = AgentSpice(model_rnn=agent_rnn._model, sindy_modules=spice_modules_optuna, n_actions=agent_rnn._n_actions)
                probs_spice = get_update_dynamics(agent=agent_spice_optuna, experiment=data_pid.xs[0])[1]
                lik_spice_after_optuna = np.exp(log_likelihood(data=data_pid.xs[0, :probs_rnn.shape[0], :agent_rnn._n_actions].numpy(), probs=probs_spice) / len(probs_spice))
                
                likelihoods_spice_after_optuna.append(np.round(lik_spice_after_optuna, 5))
                print(f"Likelihoods after optuna fitting: RNN = {np.round(lik_rnn, 5)}; SPICE =  {np.round(lik_spice_before_optuna, 5)} -> {np.round(lik_spice_after_optuna, 5)}, Diff = {np.round(lik_rnn-lik_spice_before_optuna, 5)} -> {np.round(lik_rnn-lik_spice_after_optuna, 5)}")
                
                # if lik_spice_after_optuna > lik_spice_before_optuna or np.isnan(lik_spice_before_optuna):
                sindy_modules_id = sindy_modules_optuna
                
        for rnn_module in rnn_modules:
            sindy_modules[rnn_module][pid] = sindy_modules_id[rnn_module]

    if use_optuna:
        print("Optuna-optimized participants:", optuna_pids)
        print("Likelihoods of RNN, SPICE (before), SPICE (after):")
        df = {
            'PID': optuna_pids,
            'RNN': likelihoods_rnn,
            'SPICE': likelihoods_spice_before_optuna,
            'SPICE (Optuna)': likelihoods_spice_after_optuna,
        }
        print(pd.DataFrame(data=df))
    
    # set up a SINDy-based agent by replacing the RNN-modules with the respective SINDy-model
    agent_spice = AgentSpice(model_rnn=deepcopy(agent_rnn._model), sindy_modules=sindy_modules, n_actions=agent_rnn._n_actions, deterministic=deterministic)
    
    # pruning unnecessary SINDy modules (i.e. modules which have no influence on the output but add up parameters)
    if pruning:
        agent_spice = module_pruning(
            agent_spice=agent_spice,
            dataset=data,
            participant_ids=participant_ids,
            verbose=True,
            )
    
    # remove badly fitted participants
    if filter_bad_participants:
        agent_spice, participant_ids = remove_bad_participants(
            agent_spice=agent_spice,
            agent_rnn=agent_rnn,
            dataset=data,
            participant_ids=participant_ids,
            trial_likelihood_difference_threshold=0.20,
            verbose=verbose,
        )
    
    # compute loss
    loss = None
    if get_loss and data is None:
        raise ValueError("When get_loss is True, data must be given to compute the loss. Off-policy data won't be considered to compute the loss.")
    elif get_loss and data is not None:
        loss = 0
        n_trials_total = 0
        mapping_modules_values = {module: 'x_value_choice' if 'choice' in module else 'x_value_reward' for module in agent_spice._model.submodules_sindy}
        n_parameters = agent_spice.count_parameters(mapping_modules_values=mapping_modules_values)
        for pid in participant_ids:
            mask_participant_id = data.xs[:, 0, -1] == pid
            xs, ys = data.xs[mask_participant_id].cpu().numpy(), data.ys[mask_participant_id].cpu().numpy()
            probs = get_update_dynamics(experiment=xs[0], agent=agent_spice)[1]
            loss += loss_metric(data=ys[pid, :len(probs)], probs=probs, n_parameters=n_parameters[pid])
            n_trials_total += len(probs)
        loss = loss/n_trials_total
        
    return agent_spice, loss